from mistletoe import Document, HtmlRenderer, __version__

INCLUDE = {
    "README.md": "index.html",
    "CONTRIBUTING.md": "contributing.html",
    "cutting-a-release.md": "cutting-a-release.html",
    "dev-guide.md": "dev-guide.html",
    "performance.md": "performance.html",
}

METADATA = """
<head>
  <title>mistletoe{}</title>
  <meta charset="UTF-8" />
  <meta name="description" content="A fast, extensible Markdown parser in Python." />
  <meta name="keywords" content="Markdown,Python,LaTeX,HTML" />
  <meta name="author" content="Mi Yu" />
  <meta name="viewport" content="width=device-width, initial-scale=1.0" />
  <link rel="stylesheet" href="style.css" type="text/css" />
  <link rel="stylesheet" href="https://cdnjs.cloudflare.com/ajax/libs/highlight.js/9.12.0/styles/github.min.css">
  <script src="https://cdnjs.cloudflare.com/ajax/libs/highlight.js/9.12.0/highlight.min.js"></script>
  <script>hljs.initHighlightingOnLoad();</script>
</head>
"""


class DocRenderer(HtmlRenderer):
    def render_link(self, token):
        return super().render_link(self._replace_link(token))

    def render_document(self, token, name="README.md"):
        pattern = "<html>{}<body>{}</body></html>"
        self.footnotes.update(token.footnotes)
        for filename, new_link in getattr(self, 'files', {}).items():
            for k, v in self.footnotes.items():
                if v == filename:
                    self.footnotes[k] = new_link
        subtitle = ' | {}'.format('version ' + __version__ if name == 'README.md' else name.split('.')[0].lower())
        body = '\n'.join([self.render(child) for child in token.children])
        return pattern.format(METADATA.format(subtitle), body)

    def _replace_link(self, token):
        token.target = getattr(self, 'files', {}).get(token.target, token.target)
        return token


def build(files=None):
    files = files or INCLUDE
    for f in files:
        with open(f, 'r', encoding='utf-8') as fin:
            rendered_file = 'docs/' + files[f]
            with open(rendered_file, 'w+', encoding='utf-8') as fout:
                with DocRenderer() as renderer:
                    renderer.files = files
                    print(renderer.render_document(Document(fin), f), file=fout)



from unittest import TestCase
from mistletoe import block_token
from mistletoe.block_token import Document, Heading
from mistletoe.contrib.toc_renderer import TocRenderer


class TestTocRenderer(TestCase):
    def test_parse_rendered_heading(self):
        rendered_heading = '<h3>some <em>text</em></h3>'
        content = TocRenderer.parse_rendered_heading(rendered_heading)
        self.assertEqual(content, 'some text')

    def test_render_heading(self):
        renderer = TocRenderer()
        Heading.start('### some *text*\n')
        token = Heading(Heading.read(iter(['foo'])))
        renderer.render_heading(token)
        self.assertEqual(renderer._headings[0], (3, 'some text'))

    def test_depth(self):
        renderer = TocRenderer(depth=3)
        token = Document(['# title\n', '## heading\n', '#### heading\n'])
        renderer.render(token)
        self.assertEqual(renderer._headings, [(2, 'heading')])

    def test_omit_title(self):
        renderer = TocRenderer(omit_title=True)
        token = Document(['# title\n', '\n', '## heading\n'])
        renderer.render(token)
        self.assertEqual(renderer._headings, [(2, 'heading')])

    def test_filter_conditions(self):
        import re
        filter_conds = [lambda x: re.match(r'heading', x),
                        lambda x: re.match(r'foo', x)]
        renderer = TocRenderer(filter_conds=filter_conds)
        token = Document(['# title\n',
                          '\n',
                          '## heading\n',
                          '\n',
                          '#### not heading\n'])
        renderer.render(token)
        self.assertEqual(renderer._headings, [(4, 'not heading')])

    def test_get_toc(self):
        headings = [(1, 'heading 1'),
                    (2, 'subheading 1'),
                    (2, 'subheading 2'),
                    (3, 'subsubheading 1'),
                    (2, 'subheading 3'),
                    (1, 'heading 2')]
        renderer = TocRenderer(omit_title=False)
        renderer._headings = headings
        toc = renderer.toc
        self.assertIsInstance(toc, block_token.List)
        # for now, we check at least the most nested heading
        # (hierarchy: `List -> ListItem -> {Paragraph -> RawText.content | List -> ...}`):
        heading_item = toc.children[0].children[1].children[1].children[1].children[0]
        self.assertIsInstance(heading_item, block_token.ListItem)
        self.assertEqual(heading_item.children[0].children[0].content, 'subsubheading 1')



"""
Base classes for tests.
"""

from unittest import TestCase
from mistletoe.block_token import Document


class BaseRendererTest(TestCase):
    """
    Base class for tests of renderers.
    """
    def setUp(self):
        self.maxDiff = None

    def markdownResultTest(self, markdown, expected):
        output = self.renderer.render(Document(markdown))
        self.assertEqual(output, expected)

    def filesBasedTest(func):
        """
        Note: Use this as a decorator on a test function with an empty body.
        This is a realization of the "convention over configuration"
        practice. You only need to define a unique ``sampleOutputExtension`` within your
        test case setup, in addition to the ``renderer`` under the test of course.

        Runs the current renderer against input parsed from a file and
        asserts that the renderer output is equal to content stored in another file.
        Both the "input" and "expected output" files need to have the same ``filename``
        that is extracted from the decorated test function name.
        """
        def wrapper(self):
            # take the string after the last '__' in function name
            filename = func.__name__
            filename = filename.split('__', 1)[1]

            # parse input markdown, call render on it and check the output
            with open('test/samples/{}.md'.format(filename), 'r') as fin:
                output = self.renderer.render(Document(fin))

                with open('test/samples/{}.{}'.format(filename, self.sampleOutputExtension), 'r') as expectedFin:
                    expected = ''.join(expectedFin)

                self.assertEqual(output, expected)
        return wrapper



# Copyright 2018 Tile, Inc.  All Rights Reserved.
#
# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from test.base_test import BaseRendererTest
from mistletoe.span_token import tokenize_inner
from mistletoe.contrib.jira_renderer import JiraRenderer
import random
import string

filesBasedTest = BaseRendererTest.filesBasedTest


class TestJiraRenderer(BaseRendererTest):

    def setUp(self):
        super().setUp()
        self.renderer = JiraRenderer()
        self.renderer.__enter__()
        self.addCleanup(self.renderer.__exit__, None, None, None)
        self.sampleOutputExtension = 'jira'

    def genRandomString(self, n, hasWhitespace=False):
        source = string.ascii_letters + string.digits
        if hasWhitespace:
            source = source + ' \t'

        result = ''.join(random.SystemRandom().choice(source) for _ in range(n))
        return result

    def textFormatTest(self, inputTemplate, outputTemplate):
        input = self.genRandomString(80, False)
        token = next(iter(tokenize_inner(inputTemplate.format(input))))
        output = self.renderer.render(token)
        expected = outputTemplate.format(input)
        self.assertEqual(output, expected)

    def test_escape_simple(self):
        self.textFormatTest('---fancy text---', '\\-\\-\\-fancy text\\-\\-\\-')

    def test_escape_single_chars(self):
        self.textFormatTest('**fancy \\*@\\* text**', '*fancy \\*@\\* text*')

    def test_escape_none_when_whitespaces(self):
        self.textFormatTest('obj = {{ a: (b * c) + d }}', 'obj = {{ a: (b * c) + d }}')

    def test_escape_in_inline_code(self):
        # Note: Jira puts inline code into "{{...}}" as seen in this test.
        self.textFormatTest('**code: `a = b + c;// [1]`**',
                '*code: {{{{a = b + c;// \\[1\\]}}}}*')

    def test_escape_link(self):
        # Note: There seems to be no way of how to escape plain text URL in Jira.
        self.textFormatTest('http://www.example.com', 'http://www.example.com')

    def test_render_strong(self):
        self.textFormatTest('**a{}**', '*a{}*')

    def test_render_emphasis(self):
        self.textFormatTest('*a{}*', '_a{}_')

    def test_render_inline_code(self):
        self.textFormatTest('`a{}b`', '{{{{a{}b}}}}')

    def test_render_strikethrough(self):
        self.textFormatTest('~~{}~~', '-{}-')

    def test_render_image(self):
        token = next(iter(tokenize_inner('![image](foo.jpg)')))
        output = self.renderer.render(token)
        expected = '!foo.jpg!'
        self.assertEqual(output, expected)

    def test_render_footnote_image(self):
        # token = next(tokenize_inner('![image]\n\n[image]: foo.jpg'))
        # output = self.renderer.render(token)
        # expected = '!foo.jpg!'
        # self.assertEqual(output, expected)
        pass

    def test_render_link(self):
        url = 'http://{0}.{1}.{2}'.format(self.genRandomString(5), self.genRandomString(5), self.genRandomString(3))
        body = self.genRandomString(80, True)
        token = next(iter(tokenize_inner('[{body}]({url})'.format(url=url, body=body))))
        output = self.renderer.render(token)
        expected = '[{body}|{url}]'.format(url=url, body=body)
        self.assertEqual(output, expected)

    def test_render_link_with_title(self):
        url = 'http://{0}.{1}.{2}'.format(self.genRandomString(5), self.genRandomString(5), self.genRandomString(3))
        body = self.genRandomString(80, True)
        title = self.genRandomString(20, True)
        token = next(iter(tokenize_inner('[{body}]({url} "{title}")'.format(url=url, body=body, title=title))))
        output = self.renderer.render(token)
        expected = '[{body}|{url}|{title}]'.format(url=url, body=body, title=title)
        self.assertEqual(output, expected)

    def test_render_link_with_escapes(self):
        token = next(iter(tokenize_inner('[A link with \\[brackets\\]](http://wwww.example.com/some[complex]url "[A title]")')))
        output = self.renderer.render(token)
        expected = '[A link with \\[brackets\\]|http://wwww.example.com/some\\[complex\\]url|\\[A title\\]]'
        self.assertEqual(output, expected)

    def test_render_footnote_link(self):
        pass

    def test_render_auto_link(self):
        url = 'http://{0}.{1}.{2}'.format(self.genRandomString(5), self.genRandomString(5), self.genRandomString(3))
        token = next(iter(tokenize_inner('<{url}>'.format(url=url))))
        output = self.renderer.render(token)
        expected = '[{url}]'.format(url=url)
        self.assertEqual(output, expected)

    def test_render_escape_sequence(self):
        pass

    def test_render_html_span(self):
        pass

    def test_render_heading(self):
        pass

    def test_render_quote(self):
        pass

    def test_render_paragraph(self):
        pass

    def test_render_block_code(self):
        markdown = """\
```java
public static void main(String[] args) {
    // a = 1 * 2;
}
```
"""
        expected = """\
{code:java}
public static void main(String[] args) {
    // a = 1 * 2;
}
{code}

"""
        self.markdownResultTest(markdown, expected)

    def test_render_list(self):
        pass

    def test_render_list_item(self):
        pass

    def test_render_inner(self):
        pass

    def test_render_table(self):
        pass

    def test_render_table_row(self):
        pass

    def test_render_table_cell(self):
        pass

    def test_render_thematic_break(self):
        pass

    def test_render_html_block(self):
        pass

    def test_render_document(self):
        pass

    def test_table_header(self):
        markdown = """\
| header row   |
|--------------|
| first cell   |
"""
        expected = """\
||header row||
|first cell|

"""
        self.markdownResultTest(markdown, expected)

    def test_table_empty_cell(self):
        """
        Empty cells need to have a space in them, see <https://jira.atlassian.com/browse/JRASERVER-70048>.
        """
        markdown = """\
| A | B | C |
|-----------|
| 1 |   | 3 |
"""
        expected = """\
||A||B||C||
|1| |3|

"""
        self.markdownResultTest(markdown, expected)

    def test_table_pipe_in_cell(self):
        """ Escape pipe in cells """
        markdown = """\
| A     | B |
|-----------|
| 1\\|1 | 2 |
"""
        expected = """\
||A||B||
|1\\|1|2|

"""
        self.markdownResultTest(markdown, expected)

    @filesBasedTest
    def test_render__basic_blocks(self):
        pass

    @filesBasedTest
    def test_render__lists(self):
        pass

    @filesBasedTest
    def test_render__quotes(self):
        pass



import unittest

from mistletoe import Document
from mistletoe.contrib.pygments_renderer import PygmentsRenderer
from parameterized import parameterized
from pygments.util import ClassNotFound


class TestPygmentsRenderer(unittest.TestCase):
    @parameterized.expand([(True,), (False,)])
    def test_render_no_language(self, fail_on_unsupported_language: bool):
        renderer = PygmentsRenderer(fail_on_unsupported_language=fail_on_unsupported_language)
        token = Document(['```\n', 'no language\n', '```\n'])
        output = renderer.render(token)
        expected = (
            '<div class="highlight" style="background: #f8f8f8"><pre style="line-height: 125%;">'
            '<span></span>no language\n</pre></div>\n\n'
        )
        self.assertEqual(output, expected)

    def test_render_known_language(self):
        renderer = PygmentsRenderer()
        token = Document(['```python\n', '# python language\n', '```\n'])
        output = renderer.render(token)
        expected = (
            '<div class="highlight" style="background: #f8f8f8"><pre style="line-height: 125%;">'
            '<span></span><span style="color: #3D7B7B; font-style: italic"># python language</span>\n'
            '</pre></div>\n\n'
        )
        self.assertEqual(output, expected)

    def test_render_unknown_language(self):
        renderer = PygmentsRenderer()
        token = Document(['```foobar\n', 'unknown language\n', '```\n'])
        output = renderer.render(token)
        expected = (
            '<div class="highlight" style="background: #f8f8f8"><pre style="line-height: 125%;">'
            '<span></span>unknown language\n</pre></div>\n\n'
        )
        self.assertEqual(output, expected)

    def test_render_fail_on_unsupported_language(self):
        renderer = PygmentsRenderer(fail_on_unsupported_language=True)
        token = Document(['```foobar\n', 'unknown language\n', '```\n'])
        with self.assertRaises(ClassNotFound):
            renderer.render(token)



import unittest
from unittest.mock import patch
from mistletoe import span_token


class TestBranchToken(unittest.TestCase):
    def setUp(self):
        self.addCleanup(lambda: span_token._token_types.__setitem__(-1, span_token.RawText))
        patcher = patch('mistletoe.span_token.RawText')
        self.mock = patcher.start()
        span_token._token_types[-1] = self.mock
        self.addCleanup(patcher.stop)

    def _test_parse(self, token_cls, raw, arg, **kwargs):
        token = next(iter(span_token.tokenize_inner(raw)))
        self.assertIsInstance(token, token_cls)
        self._test_token(token, arg, **kwargs)
        return token

    def _test_token(self, token, arg, children=True, **kwargs):
        for attr, value in kwargs.items():
            self.assertEqual(getattr(token, attr), value)
        if children:
            self.mock.assert_any_call(arg)


class TestStrong(TestBranchToken):
    def test_parse(self):
        self._test_parse(span_token.Strong, '**some text**', 'some text')
        self._test_parse(span_token.Strong, '__some text__', 'some text')

    def test_strong_when_both_delimiter_run_lengths_are_multiples_of_3(self):
        tokens = iter(span_token.tokenize_inner('foo******bar*********baz'))
        self._test_token(next(tokens), 'foo', children=False)
        self._test_token(next(tokens), 'bar', children=True)
        self._test_token(next(tokens), '***baz', children=False)

    def test_overlapping_delimiter_runs_do_not_crash(self):
        tokens = list(span_token.tokenize_inner('**a****a*'))
        self.assertEqual(len(tokens), 3)
        self.assertIsInstance(tokens[0], span_token.Strong)
        self.assertIsInstance(tokens[2], span_token.Emphasis)
        self.mock.assert_any_call('*')

    def test_issue_261_asterisk_sequences_do_not_crash(self):
        cases = [
            '**IKXZXPTB****AJFHV *Pbjovp*VSN*',
            '**IKXZXPTB****AJFHV*Pbjovp*',
            'Ch******her R**a, Bu*****ly In****nt, j**n ko***ob, '
            'ji*yu, M**s, Ch****us, M*',
        ]
        for case in cases:
            with self.subTest(case=case):
                self.assertTrue(list(span_token.tokenize_inner(case)))


class TestEmphasis(TestBranchToken):
    def test_parse(self):
        self._test_parse(span_token.Emphasis, '*some text*', 'some text')
        self._test_parse(span_token.Emphasis, '_some text_', 'some text')

    def test_emphasis_with_straight_quote(self):
        tokens = iter(span_token.tokenize_inner('_Book Title_\'s author'))
        self._test_token(next(tokens), 'Book Title', children=True)
        self._test_token(next(tokens), '\'s author', children=False)

    def test_emphasis_with_smart_quote(self):
        tokens = iter(span_token.tokenize_inner('_Book Title_’s author'))
        self._test_token(next(tokens), 'Book Title', children=True)
        self._test_token(next(tokens), '’s author', children=False)

    def test_no_emphasis_for_underscore_without_punctuation(self):
        tokens = iter(span_token.tokenize_inner('_an example without_punctuation'))
        self._test_token(next(tokens), '_an example without_punctuation', children=True)

    def test_emphasis_for_asterisk_without_punctuation(self):
        tokens = iter(span_token.tokenize_inner('*an example without*punctuation'))
        self._test_token(next(tokens), 'an example without', children=True)
        self._test_token(next(tokens), 'punctuation', children=False)


class TestInlineCode(TestBranchToken):
    def _test_parse_enclosed(self, encl_type, encl_delimiter):
        token = self._test_parse(encl_type, '{delim}`some text`{delim}'.format(delim=encl_delimiter), 'some text')
        self.assertEqual(len(token.children), 1)
        self.assertIsInstance(token.children[0], span_token.InlineCode)

    def test_parse(self):
        self._test_parse(span_token.InlineCode, '`some text`', 'some text')

    def test_parse_in_bold(self):
        self._test_parse_enclosed(span_token.Strong, '**')
        self._test_parse_enclosed(span_token.Strong, '__')

    def test_parse_in_emphasis(self):
        self._test_parse_enclosed(span_token.Emphasis, '*')
        self._test_parse_enclosed(span_token.Emphasis, '_')

    def test_parse_in_strikethrough(self):
        self._test_parse_enclosed(span_token.Strikethrough, '~~')

    def test_remove_space_if_present_on_both_sides(self):
        self._test_parse(span_token.InlineCode, '``` ```', ' ')
        self._test_parse(span_token.InlineCode, '`  ``  `', ' `` ')

    def test_preserve_escapes(self):
        self._test_parse(span_token.InlineCode, '`\\xa0b\\xa0`', '\\xa0b\\xa0')
        self._test_parse(span_token.InlineCode, '``\\`\\[``', '\\`\\[')


class TestStrikethrough(TestBranchToken):
    def test_parse(self):
        self._test_parse(span_token.Strikethrough, '~~some text~~', 'some text')

    def test_parse_multiple(self):
        tokens = iter(span_token.tokenize_inner('~~one~~ ~~two~~'))
        self._test_token(next(tokens), 'one')
        self._test_token(next(tokens), 'two')


class TestLink(TestBranchToken):
    def test_parse(self):
        self._test_parse(span_token.Link, '[name 1](target1)', 'name 1',
                         target='target1', title='')

    def test_parse_multi_links(self):
        tokens = iter(span_token.tokenize_inner('[n1](t1) & [n2](t2)'))
        self._test_token(next(tokens), 'n1', target='t1')
        self._test_token(next(tokens), ' & ', children=False)
        self._test_token(next(tokens), 'n2', target='t2')

    def test_parse_children(self):
        token = next(iter(span_token.tokenize_inner('[![alt](src)](target)')))
        child = next(iter(token.children))
        self._test_token(child, 'alt', src='src')

    def test_parse_angle_bracketed_inline_link_with_space(self):
        self._test_parse(span_token.Link, '[link](</my uri> \'a title\')',
            'link', target='/my uri', title='a title')


class TestAutoLink(TestBranchToken):
    def test_parse(self):
        self._test_parse(span_token.AutoLink, '<ftp://foo.com>', 'ftp://foo.com', target='ftp://foo.com')


class TestImage(TestBranchToken):
    def test_parse(self):
        self._test_parse(span_token.Image, '![alt](link)', 'alt', src='link')
        self._test_parse(span_token.Image, '![alt](link "title")', 'alt',
                         src='link', title='title')

    def test_no_alternative_text(self):
        self._test_parse(span_token.Image, '![](link)', '', children=False, src='link')


class TestEscapeSequence(TestBranchToken):
    def test_parse(self):
        self._test_parse(span_token.EscapeSequence, r'\*', '*')

    def test_parse_in_text(self):
        tokens = iter(span_token.tokenize_inner(r'some \*text*'))
        self._test_token(next(tokens), 'some ', children=False)
        self._test_token(next(tokens), '*')
        self._test_token(next(tokens), 'text*', children=False)


class TestRawText(unittest.TestCase):
    def test_attribute(self):
        token = span_token.RawText('some text')
        self.assertEqual(token.content, 'some text')

    def test_no_children(self):
        token = span_token.RawText('some text')
        self.assertIsNone(token.children)

    def test_valid_html_entities(self):
        tokens = span_token.tokenize_inner('&nbsp; &#21512;')
        self.assertEqual(tokens[0].content, '\xa0 \u5408')

    def test_invalid_html_entities(self):
        text = '&nbsp &x; &#; &#x; &#87654321; &#abcdef0; &ThisIsNotDefined; &hi?;'
        tokens = span_token.tokenize_inner(text)
        self.assertEqual(tokens[0].content, text)


class TestLineBreak(unittest.TestCase):
    def test_parse_soft_break(self):
        token, = span_token.tokenize_inner('\n')
        self.assertIsInstance(token, span_token.LineBreak)
        self.assertTrue(token.soft)

    def test_parse_hard_break_with_double_blanks(self):
        token, = span_token.tokenize_inner('  \n')
        self.assertIsInstance(token, span_token.LineBreak)
        self.assertFalse(token.soft)

    def test_parse_hard_break_with_backslash(self):
        _, token, = span_token.tokenize_inner(' \\\n')
        self.assertIsInstance(token, span_token.LineBreak)
        self.assertFalse(token.soft)


class TestContains(unittest.TestCase):
    def test_contains(self):
        token = next(iter(span_token.tokenize_inner('**with some *emphasis* text**')))
        self.assertTrue('text' in token)
        self.assertTrue('emphasis' in token)
        self.assertFalse('foo' in token)


class TestParent(unittest.TestCase):
    def test_parent(self):
        token, = span_token.tokenize_inner('**some text**')
        self.assertIsInstance(token.children[0], span_token.RawText)
        self.assertEqual(token.children[0].parent, token)


class TestHtmlSpan(unittest.TestCase):
    def setUp(self):
        span_token.add_token(span_token.HtmlSpan)
        self.addCleanup(span_token.reset_tokens)

    def test_parse(self):
        tokens = span_token.tokenize_inner('<a>')
        self.assertIsInstance(tokens[0], span_token.HtmlSpan)
        self.assertEqual('<a>', tokens[0].content)

    def test_parse_with_illegal_whitespace(self):
        tokens = span_token.tokenize_inner('< a><\nfoo><bar/ >\n<foo bar=baz\nbim!bop />')
        for t in tokens:
            self.assertNotIsInstance(t, span_token.HtmlSpan)



import unittest

import mistletoe.block_token as block_token
import mistletoe.span_token as span_token
from mistletoe.markdown_renderer import (
    LinkReferenceDefinition,
    LinkReferenceDefinitionBlock,
)


class TestLineNumbers(unittest.TestCase):
    def setUp(self) -> None:
        block_token.add_token(block_token.HTMLBlock)
        span_token.add_token(span_token.HTMLSpan)
        block_token.remove_token(block_token.Footnote)
        block_token.add_token(LinkReferenceDefinitionBlock)
        return super().setUp()

    def tearDown(self) -> None:
        span_token.reset_tokens()
        block_token.reset_tokens()
        return super().tearDown()

    def test_main(self):
        # see line_numbers.md for a description of how the test works.
        NUMBER_OF_LINE_NUMBERS_TO_BE_CHECKED = 13
        with open("test/samples/line_numbers.md", "r") as fin:
            document = block_token.Document(fin)
        count = self.check_line_numbers(document)
        self.assertEqual(count, NUMBER_OF_LINE_NUMBERS_TO_BE_CHECKED)

    def check_line_numbers(self, token: block_token.BlockToken):
        """Check the line number on the given block token and its children, if possible."""
        count = 0
        line_number = self.get_expected_line_number(token)
        if line_number:
            self.assertEqual(token.line_number, line_number)
            count += 1

        if isinstance(token, block_token.Table):
            count += self.check_line_numbers(token.header)

        for child in token.children:
            if isinstance(child, block_token.BlockToken):
                count += self.check_line_numbers(child)

        return count

    def get_expected_line_number(self, token: block_token.BlockToken):
        # the expected line number, if it exists, should be wrapped in an inline
        # code token and be an immediate child of the token.
        # or it could be the title of a link reference definition.
        for child in token.children:
            if isinstance(child, span_token.InlineCode):
                return int(child.children[0].content)
            if isinstance(child, LinkReferenceDefinition):
                return int(child.title)



import unittest
from mistletoe import Document, ast_renderer


class TestAstRenderer(unittest.TestCase):
    def test(self):
        self.maxDiff = None
        d = Document([
            '# heading 1\n',
            '\n',
            'hello\n',
            'world\n',
        ])
        output = ast_renderer.get_ast(d)
        expected = {'type': 'Document',
                  'footnotes': {},
                  'line_number': 1,
                  'children': [{
                      'type': 'Heading',
                      'level': 1,
                      'line_number': 1,
                      'children': [{
                          'type': 'RawText',
                          'content': 'heading 1'
                      }]
                  }, {
                      'type': 'Paragraph',
                      'line_number': 3,
                      'children': [{
                          'type': 'RawText',
                          'content': 'hello'
                      }, {
                          'type': 'LineBreak',
                          'soft': True,
                          'content': ''
                      }, {
                          'type': 'RawText',
                          'content': 'world'
                      }]
                  }]}
        self.assertEqual(output, expected)

    def test_footnotes(self):
        self.maxDiff = None
        d = Document([
            '[bar][baz]\n',
            '\n',
            '[baz]: spam\n',
        ])
        expected = {'type': 'Document',
                  'footnotes': {'baz': ('spam', '')},
                  'line_number': 1,
                  'children': [{
                      'type': 'Paragraph',
                      'line_number': 1,
                      'children': [{
                          'type': 'Link',
                          'target': 'spam',
                          'title': '',
                          'children': [{
                              'type': 'RawText',
                              'content': 'bar'
                          }]
                      }]
                  }]}
        output = ast_renderer.get_ast(d)
        self.assertEqual(output, expected)

    def test_table(self):
        self.maxDiff = None
        d = Document([
            "| A   | B   |\n",
            "| --- | --- |\n",
            "| 1   | 2   |\n",
        ])
        expected = {
            "type": "Document",
            "footnotes": {},
            'line_number': 1,
            "children": [{
                "type": "Table",
                "column_align": [None, None],
                'line_number': 1,
                "header": {
                    "type": "TableRow",
                    "row_align": [None, None],
                    'line_number': 1,
                    "children": [{
                        "type": "TableCell",
                        "align": None,
                        'line_number': 1,
                        "children": [{
                            "type": "RawText",
                            "content": "A",
                        }]}, {
                        "type": "TableCell",
                        "align": None,
                        'line_number': 1,
                        "children": [{
                            "type": "RawText",
                            "content": "B",
                        }]
                    }],
                },
                "children": [{
                    "type": "TableRow",
                    "row_align": [None, None],
                    'line_number': 3,
                    "children": [{
                        "type": "TableCell",
                        "align": None,
                        'line_number': 3,
                        "children": [{
                            "type": "RawText",
                            "content": "1",
                        }]}, {
                        "type": "TableCell",
                        "align": None,
                        'line_number': 3,
                        "children": [{
                            "type": "RawText",
                            "content": "2",
                        }]
                    }],
                }],
            }],
        }
        output = ast_renderer.get_ast(d)
        self.assertEqual(output, expected)



from unittest import TestCase, mock
from mistletoe import Document
from mistletoe.html_renderer import HtmlRenderer
from parameterized import parameterized


class TestRenderer(TestCase):
    def setUp(self):
        self.renderer = HtmlRenderer()
        self.renderer.render_inner = mock.Mock(return_value='inner')
        self.renderer.__enter__()
        self.addCleanup(self.renderer.__exit__, None, None, None)

    def _test_token(self, token_name, expected_output, children=True,
                    without_attrs=None, **kwargs):
        render_func = self.renderer.render_map[token_name]
        children = mock.MagicMock(spec=list) if children else None
        mock_token = mock.Mock(children=children, **kwargs)
        without_attrs = without_attrs or []
        for attr in without_attrs:
            delattr(mock_token, attr)
        self.assertEqual(render_func(mock_token), expected_output)


class TestHtmlRenderer(TestRenderer):
    def test_strong(self):
        self._test_token('Strong', '<strong>inner</strong>')

    def test_emphasis(self):
        self._test_token('Emphasis', '<em>inner</em>')

    def test_inline_code(self):
        from mistletoe.span_token import tokenize_inner
        output = self.renderer.render(tokenize_inner('`foo`')[0])
        self.assertEqual(output, '<code>foo</code>')
        output = self.renderer.render(tokenize_inner('`` \\[\\` ``')[0])
        self.assertEqual(output, '<code>\\[\\`</code>')

    def test_strikethrough(self):
        self._test_token('Strikethrough', '<del>inner</del>')

    def test_image(self):
        expected = '<img src="s&amp;r%3Ec" alt="" title="title" />'
        self._test_token('Image', expected, src='s&r>c', title='title')

    def test_link(self):
        expected = '<a href="target" title="title">inner</a>'
        self._test_token('Link', expected, target='target', title='title')

    def test_autolink(self):
        expected = '<a href="link">inner</a>'
        self._test_token('AutoLink', expected, target='link', mailto=False)

    def test_email_autolink_escapes_href(self):
        expected = '<a href="mailto:a&amp;b@example.com">inner</a>'
        self._test_token('AutoLink', expected, target='a&b@example.com',
                         mailto=True)

    def test_escape_sequence(self):
        self._test_token('EscapeSequence', 'inner')

    def test_raw_text(self):
        self._test_token('RawText', 'john &amp; jane',
                         children=False, content='john & jane')

    def test_html_span(self):
        self._test_token('HtmlSpan', '<some>text</some>',
                         children=False, content='<some>text</some>')

    def test_heading(self):
        expected = '<h3>inner</h3>'
        self._test_token('Heading', expected, level=3)

    def test_quote(self):
        expected = '<blockquote>\n</blockquote>'
        self._test_token('Quote', expected)

    def test_paragraph(self):
        self._test_token('Paragraph', '<p>inner</p>')

    def test_block_code(self):
        from mistletoe.block_token import tokenize
        output = self.renderer.render(tokenize(['```sh\n', 'foo\n', '```\n'])[0])
        expected = '<pre><code class="language-sh">foo\n</code></pre>'
        self.assertEqual(output, expected)

    def test_block_code_no_language(self):
        from mistletoe.block_token import tokenize
        output = self.renderer.render(tokenize(['```\n', 'foo\n', '```\n'])[0])
        expected = '<pre><code>foo\n</code></pre>'
        self.assertEqual(output, expected)

    def test_list(self):
        expected = '<ul>\n\n</ul>'
        self._test_token('List', expected, start=None)

    def test_list_item(self):
        expected = '<li></li>'
        self._test_token('ListItem', expected)

    def test_table_with_header(self):
        func_path = 'mistletoe.html_renderer.HtmlRenderer.render_table_row'
        with mock.patch(func_path, autospec=True) as mock_func:
            mock_func.return_value = 'row'
            expected = ('<table>\n'
                        '<thead>\nrow</thead>\n'
                        '<tbody>\ninner</tbody>\n'
                      '</table>')
            self._test_token('Table', expected)

    def test_table_without_header(self):
        func_path = 'mistletoe.html_renderer.HtmlRenderer.render_table_row'
        with mock.patch(func_path, autospec=True) as mock_func:
            mock_func.return_value = 'row'
            expected = '<table>\n<tbody>\ninner</tbody>\n</table>'
            self._test_token('Table', expected, without_attrs=['header',])

    def test_table_row(self):
        self._test_token('TableRow', '<tr>\n</tr>\n')

    def test_table_cell(self):
        expected = '<td align="left">inner</td>\n'
        self._test_token('TableCell', expected, align=None)

    def test_table_cell0(self):
        expected = '<td align="center">inner</td>\n'
        self._test_token('TableCell', expected, align=0)

    def test_table_cell1(self):
        expected = '<td align="right">inner</td>\n'
        self._test_token('TableCell', expected, align=1)

    def test_thematic_break(self):
        self._test_token('ThematicBreak', '<hr />', children=False)

    def test_html_block(self):
        content = expected = '<h1>hello</h1>\n<p>this is\na paragraph</p>\n'
        self._test_token('HtmlBlock', expected,
                         children=False, content=content)

    def test_line_break(self):
        self._test_token('LineBreak', '<br />\n', children=False, soft=False)

    def test_document(self):
        self._test_token('Document', '', footnotes={})


class TestHtmlRendererEscaping(TestCase):
    @parameterized.expand([
        (False, False, '" and \''),
        (False, True, '" and &#x27;'),
        (True, False, '&quot; and \''),
        (True, True, '&quot; and &#x27;'),
    ])
    def test_escape_html_text(self, escape_double, escape_single, expected):
        with HtmlRenderer(html_escape_double_quotes=escape_double,
                          html_escape_single_quotes=escape_single) as renderer:
            self.assertEqual(renderer.escape_html_text('" and \''), expected)

    def test_unprocessed_html_tokens_escaped(self):
        with HtmlRenderer(process_html_tokens=False) as renderer:
            token = Document(['<div><br> as plain text</div>\n'])
            expected = '<p>&lt;div&gt;&lt;br&gt; as plain text&lt;/div&gt;</p>\n'
            self.assertEqual(renderer.render(token), expected)

    def test_email_autolink_escapes_href(self):
        with HtmlRenderer() as renderer:
            output = renderer.render(Document('<a&b@example.com>'))
        expected = '<p><a href="mailto:a&amp;b@example.com">a&amp;b@example.com</a></p>\n'
        self.assertEqual(output, expected)


class TestHtmlRendererFootnotes(TestCase):
    def setUp(self):
        self.renderer = HtmlRenderer()
        self.renderer.__enter__()
        self.addCleanup(self.renderer.__exit__, None, None, None)

    def test_footnote_image(self):
        token = Document(['![alt][foo]\n', '\n', '[foo]: bar "title"\n'])
        expected = '<p><img src="bar" alt="alt" title="title" /></p>\n'
        self.assertEqual(self.renderer.render(token), expected)

    def test_footnote_link(self):
        token = Document(['[name][foo]\n', '\n', '[foo]: target\n'])
        expected = '<p><a href="target">name</a></p>\n'
        self.assertEqual(self.renderer.render(token), expected)



import unittest
from unittest.mock import call, patch

from parameterized import parameterized

from mistletoe import block_token, block_tokenizer, span_token


class TestToken(unittest.TestCase):
    def setUp(self):
        self.addCleanup(lambda: span_token._token_types.__setitem__(-1, span_token.RawText))
        patcher = patch('mistletoe.span_token.RawText')
        self.mock = patcher.start()
        span_token._token_types[-1] = self.mock
        self.addCleanup(patcher.stop)

    def _test_match(self, token_cls, lines, arg, **kwargs):
        token = next(iter(block_token.tokenize(lines)))
        self.assertIsInstance(token, token_cls)
        self._test_token(token, arg, **kwargs)

    def _test_token(self, token, arg, **kwargs):
        for attr, value in kwargs.items():
            self.assertEqual(getattr(token, attr), value)
        self.mock.assert_any_call(arg)


class TestAtxHeading(TestToken):
    def test_match(self):
        lines = ['### heading 3\n']
        arg = 'heading 3'
        self._test_match(block_token.Heading, lines, arg, level=3)

    def test_children_with_enclosing_hashes(self):
        lines = ['# heading 3 #####  \n']
        arg = 'heading 3'
        self._test_match(block_token.Heading, lines, arg, level=1)

    def test_not_heading(self):
        lines = ['####### paragraph\n']
        arg = '####### paragraph'
        self._test_match(block_token.Paragraph, lines, arg)

    def test_heading_in_paragraph(self):
        lines = ['foo\n', '# heading\n', 'bar\n']
        token1, token2, token3 = block_token.tokenize(lines)
        self.assertIsInstance(token1, block_token.Paragraph)
        self.assertIsInstance(token2, block_token.Heading)
        self.assertIsInstance(token3, block_token.Paragraph)


class TestSetextHeading(TestToken):
    def test_match(self):
        lines = ['some heading\n', '---\n']
        arg = 'some heading'
        self._test_match(block_token.SetextHeading, lines, arg, level=2)

    def test_next(self):
        lines = ['some\n', 'heading\n', '---\n', '\n', 'foobar\n']
        tokens = iter(block_token.tokenize(lines))
        self.assertIsInstance(next(tokens), block_token.SetextHeading)
        self.assertIsInstance(next(tokens), block_token.Paragraph)
        self.mock.assert_has_calls([call('some'), call('heading'), call('foobar')])
        with self.assertRaises(StopIteration):
            next(tokens)


class TestQuote(unittest.TestCase):
    def test_match(self):
        with patch('mistletoe.block_token.Paragraph'):
            token = next(iter(block_token.tokenize(['> line 1\n', '> line 2\n'])))
            self.assertIsInstance(token, block_token.Quote)

    def test_lazy_continuation(self):
        with patch('mistletoe.block_token.Paragraph'):
            token = next(iter(block_token.tokenize(['> line 1\n', 'line 2\n'])))
            self.assertIsInstance(token, block_token.Quote)


class TestCodeFence(TestToken):
    def test_match_fenced_code(self):
        lines = ['```sh\n', 'rm dir\n', 'mkdir test\n', '```\n']
        arg = 'rm dir\nmkdir test\n'
        self._test_match(block_token.CodeFence, lines, arg, language='sh')

    def test_match_fenced_code_with_tilde(self):
        lines = ['~~~sh\n', 'rm dir\n', 'mkdir test\n', '~~~\n']
        arg = 'rm dir\nmkdir test\n'
        self._test_match(block_token.CodeFence, lines, arg, language='sh')

    def test_not_match_fenced_code_when_only_inline_code(self):
        lines = ['`~` is called tilde']
        token = next(iter(block_token.tokenize(lines)))
        self.assertIsInstance(token, block_token.Paragraph)
        token1 = token.children[0]
        self.assertIsInstance(token1, span_token.InlineCode)
        self.mock.assert_has_calls([call('~'), call(' is called tilde')])

    def test_mixed_code_fence(self):
        lines = ['~~~markdown\n', '```sh\n', 'some code\n', '```\n', '~~~\n']
        arg = '```sh\nsome code\n```\n'
        self._test_match(block_token.CodeFence, lines, arg, language='markdown')

    def test_fence_code_lazy_continuation(self):
        lines = ['```sh\n', 'rm dir\n', '\n', 'mkdir test\n', '```\n']
        arg = 'rm dir\n\nmkdir test\n'
        self._test_match(block_token.CodeFence, lines, arg, language='sh')

    def test_no_wrapping_newlines_code_fence(self):
        lines = ['```\n', 'hey', '```\n', 'paragraph\n']
        arg = 'hey'
        self._test_match(block_token.CodeFence, lines, arg, language='')

    def test_unclosed_code_fence(self):
        lines = ['```\n', 'hey']
        arg = 'hey'
        self._test_match(block_token.CodeFence, lines, arg, language='')

    def test_code_fence_with_backticks_and_tildes_in_the_info_string(self):
        lines = ['~~~ aa ``` ~~~\n', 'foo\n', '~~~\n']
        arg = 'foo\n'
        self._test_match(block_token.CodeFence, lines, arg, language='aa')


class TestBlockCode(TestToken):
    def test_parse_indented_code(self):
        lines = ['    rm dir\n', '    mkdir test\n']
        arg = 'rm dir\nmkdir test\n'
        self._test_match(block_token.BlockCode, lines, arg, language='')

    def test_parse_indented_code_with_blank_lines(self):
        lines = ['    chunk1\n', '\n', '    chunk2\n', '  \n', ' \n', ' \n', '    chunk3\n']
        arg = 'chunk1\n\nchunk2\n\n\n\nchunk3\n'
        self._test_match(block_token.BlockCode, lines, arg, language='')


class TestParagraph(TestToken):
    def setUp(self):
        super().setUp()
        block_token.add_token(block_token.HtmlBlock)
        self.addCleanup(block_token.reset_tokens)

    def test_parse(self):
        lines = ['some\n', 'continuous\n', 'lines\n']
        arg = 'some'
        self._test_match(block_token.Paragraph, lines, arg)

    def test_read(self):
        lines = ['this\n', '```\n', 'is some\n', '```\n', 'code\n']
        try:
            token1, token2, token3 = block_token.tokenize(lines)
        except ValueError as e:
            raise AssertionError("Token number mismatch.") from e
        self.assertIsInstance(token1, block_token.Paragraph)
        self.assertIsInstance(token2, block_token.CodeFence)
        self.assertIsInstance(token3, block_token.Paragraph)

    def test_parse_interrupting_block_tokens(self):
        interrupting_blocks = [
            '***\n',  # thematic break
            '## atx\n',  # ATX heading
            '<div>\n',  # HTML block type 6
            '> block quote\n',
            '1. list\n',
            '``` fenced code block\n',
            ('| table |\n', '| ----- |\n', '| row   |\n'),
        ]
        for block in interrupting_blocks:
            lines = ['Paragraph 1\n', *block]
        try:
            token1, token2 = block_token.tokenize(lines)
        except ValueError as e:
            raise AssertionError("Token number mismatch. Lines: '{}'".format(lines)) from e
        self.assertIsInstance(token1, block_token.Paragraph)
        self.assertNotIsInstance(token2, block_token.Paragraph)

    def test_parse_non_interrupting_block_tokens(self):
        lines = [
            'Paragraph 1\n',
            '2. list\n',  # list doesn't start from 1
            '    indented text\n',  # code block
            '<custom>\n',  # HTML block type 7
            '\n',
            'Paragraph 2\n'
        ]
        try:
            token1, token2 = block_token.tokenize(lines)
        except ValueError as e:
            raise AssertionError("Token number mismatch.") from e
        self.assertIsInstance(token1, block_token.Paragraph)
        self.assertIsInstance(token2, block_token.Paragraph)

    def test_parse_setext_heading(self):
        lines = [
            'Two line\n',
            'heading\n',
            '---\n'
        ]
        try:
            token1, = block_token.tokenize(lines)
        except ValueError as e:
            raise AssertionError("Token number mismatch.") from e
        self.assertIsInstance(token1, block_token.SetextHeading)


class TestListItem(unittest.TestCase):
    def test_parse_marker(self):
        lines = ['- foo\n',
                 '   *    bar\n',
                 ' + baz\n',
                 '1. item 1\n',
                 '2) item 2\n',
                 '123456789. item x\n',
                 '*\n']
        for line in lines:
            self.assertTrue(block_token.ListItem.parse_marker(line))
        bad_lines = ['> foo\n',
                     '1item 1\n',
                     '2| item 2\n',
                     '1234567890. item x\n',
                     '    * too many spaces\n']
        for line in bad_lines:
            self.assertFalse(block_token.ListItem.parse_marker(line))

    def test_tokenize(self):
        lines = [' -    foo\n',
                 '   bar\n',
                 '\n',
                 '          baz\n']
        token1, token2 = next(iter(block_token.tokenize(lines))).children[0].children
        self.assertIsInstance(token1, block_token.Paragraph)
        self.assertTrue('foo' in token1)
        self.assertIsInstance(token2, block_token.BlockCode)

    def test_sublist(self):
        lines = ['- foo\n',
                 '  - bar\n']
        token1, token2 = block_token.tokenize(lines)[0].children[0].children
        self.assertIsInstance(token1, block_token.Paragraph)
        self.assertIsInstance(token2, block_token.List)

    def test_deep_list(self):
        lines = ['- foo\n',
                 '  - bar\n',
                 '    - baz\n']
        ptoken, ltoken = block_token.tokenize(lines)[0].children[0].children
        self.assertIsInstance(ptoken, block_token.Paragraph)
        self.assertIsInstance(ltoken, block_token.List)
        self.assertTrue('foo' in ptoken)
        ptoken, ltoken = ltoken.children[0].children
        self.assertIsInstance(ptoken, block_token.Paragraph)
        self.assertTrue('bar' in ptoken)
        self.assertIsInstance(ltoken, block_token.List)
        self.assertTrue('baz' in ltoken)

    def test_loose_list(self):
        lines = ['- foo\n',
                 '  ~~~\n',
                 '  bar\n',
                 '  \n',
                 '  baz\n'
                 '  ~~~\n']
        list_item = block_token.tokenize(lines)[0].children[0]
        self.assertEqual(list_item.loose, False)

    def test_tight_list(self):
        lines = ['- foo\n',
                 '\n',
                 '# bar\n']
        list_item = block_token.tokenize(lines)[0].children[0]
        self.assertEqual(list_item.loose, False)

    def test_tabbed_list_items(self):
        # according to the CommonMark spec:
        # in contexts where spaces help to define block structure, tabs behave as if they
        # were replaced by spaces with a tab stop of 4 characters.
        lines = ['title\n',
                 '*\ttabbed item long line\n',
                 '\n',  # break lazy continuation
                 '    continuation 1\n',
                 '*   second list item\n',
                 '\n',  # break lazy continuation
                 '\tcontinuation 2\n']
        tokens = block_token.tokenize(lines)
        self.assertEqual(len(tokens), 2)
        self.assertIsInstance(tokens[0], block_token.Paragraph)
        self.assertIsInstance(tokens[1], block_token.List)
        self.assertTrue('tabbed item long line' in tokens[1].children[0])
        self.assertTrue('continuation 1' in tokens[1].children[0])
        self.assertTrue('second list item' in tokens[1].children[1])
        self.assertTrue('continuation 2' in tokens[1].children[1])

    def test_list_items_starting_with_blank_line(self):
        lines = ['-\n',
                 '  foo\n',
                 '-\n',
                 '  ```\n',
                 '  bar\n',
                 '  ```\n',
                 '-\n',
                 '      baz\n']
        tokens = block_token.tokenize(lines)
        self.assertEqual(len(tokens), 1)
        self.assertIsInstance(tokens[0], block_token.List)
        self.assertIsInstance(tokens[0].children[0].children[0], block_token.Paragraph)
        self.assertIsInstance(tokens[0].children[1].children[0], block_token.CodeFence)
        self.assertIsInstance(tokens[0].children[2].children[0], block_token.BlockCode)
        self.assertTrue('foo' in tokens[0].children[0].children[0])
        self.assertEqual('bar\n', tokens[0].children[1].children[0].children[0].content)
        self.assertEqual('baz\n', tokens[0].children[2].children[0].children[0].content)

    def test_a_list_item_may_begin_with_at_most_one_blank_line(self):
        lines = ['-\n',
                 '\n',
                 '  foo\n']
        tokens = block_token.tokenize(lines)
        self.assertEqual(len(tokens), 2)
        self.assertIsInstance(tokens[0], block_token.List)
        self.assertIsInstance(tokens[1], block_token.Paragraph)
        self.assertTrue('foo' in tokens[1].children[0])

    def test_empty_list_item_in_the_middle(self):
        lines = ['* a\n',
                 '*\n',
                 '\n',
                 '* c\n']
        tokens = block_token.tokenize(lines)
        self.assertEqual(len(tokens), 1)
        self.assertIsInstance(tokens[0], block_token.List)
        self.assertEqual(len(tokens[0].children), 3)
        self.assertTrue(tokens[0].loose)

    def test_list_with_code_block(self):
        lines = ['1.      indented code\n',
                 '\n',
                 '   paragraph\n',
                 '\n',
                 '       more code\n']
        tokens = block_token.tokenize(lines)
        self.assertEqual(len(tokens), 1)
        self.assertIsInstance(tokens[0], block_token.List)
        self.assertEqual(len(tokens[0].children), 1)
        self.assertIsInstance(tokens[0].children[0].children[0], block_token.BlockCode)
        self.assertEqual(' indented code\n', tokens[0].children[0].children[0].children[0].content)
        self.assertIsInstance(tokens[0].children[0].children[1], block_token.Paragraph)
        self.assertIsInstance(tokens[0].children[0].children[2], block_token.BlockCode)


class TestList(unittest.TestCase):
    def test_different_markers(self):
        lines = ['- foo\n',
                 '* bar\n',
                 '1. baz\n',
                 '2) spam\n']
        l1, l2, l3, l4 = block_token.tokenize(lines)
        self.assertIsInstance(l1, block_token.List)
        self.assertTrue('foo' in l1)
        self.assertIsInstance(l2, block_token.List)
        self.assertTrue('bar' in l2)
        self.assertIsInstance(l3, block_token.List)
        self.assertTrue('baz' in l3)
        self.assertIsInstance(l4, block_token.List)
        self.assertTrue('spam' in l4)

    def test_sublist(self):
        lines = ['- foo\n',
                 '  + bar\n']
        token, = block_token.tokenize(lines)
        self.assertIsInstance(token, block_token.List)


class TestTable(unittest.TestCase):
    def test_parse_align(self):
        test_func = block_token.Table.parse_align
        self.assertEqual(test_func(':------'), None)
        self.assertEqual(test_func(':-----:'), 0)
        self.assertEqual(test_func('------:'), 1)

    def test_parse_delimiter(self):
        def test_func(s):
            return block_token.Table.split_delimiter(s)
        self.assertEqual(list(test_func('|-| :--- | :---: | ---:|\n')),
                ['-', ':---', ':---:', '---:'])

    @parameterized.expand([
        ('| --- | --- | --- |\n'),
        ('| - | - | - |\n'),
        ('|-|-|-- \n'),
    ])
    def test_match(self, delimiter_line):
        lines = ['| header 1 | header 2 | header 3 |\n',
                delimiter_line,
                 '| cell 1 | cell 2 | cell 3 |\n',
                 '| more 1 | more 2 | more 3 |\n']
        with patch('mistletoe.block_token.TableRow') as mock:
            token, = block_token.tokenize(lines)
            self.assertIsInstance(token, block_token.Table)
            self.assertTrue(hasattr(token, 'header'))
            self.assertEqual(token.column_align, [None, None, None])
            token.children
            calls = [call(line, [None, None, None], line_number) for line_number, line in enumerate(lines, start=1) if line_number != 2]
            mock.assert_has_calls(calls)

    def test_easy_table(self):
        lines = ['header 1 | header 2\n',
                 '    ---: | :---\n',
                 '  cell 1 | cell 2\n']
        with patch('mistletoe.block_token.TableRow') as mock:
            token, = block_token.tokenize(lines)
            self.assertIsInstance(token, block_token.Table)
            self.assertTrue(hasattr(token, 'header'))
            self.assertEqual(token.column_align, [1, None])
            token.children
            calls = [call(line, [1, None], line_number) for line_number, line in enumerate(lines, start=1) if line_number != 2]
            mock.assert_has_calls(calls)

    def test_not_easy_table(self):
        lines = ['not header 1 | not header 2\n',
                 'foo | bar\n']
        token, = block_token.tokenize(lines)
        self.assertIsInstance(token, block_token.Paragraph)

    def test_interrupt_paragraph_option(self):
        lines = [
            'Paragraph 1\n',
            '| table |\n',
            '| ----- |\n',
            '| row   |\n',
        ]
        try:
            block_token.Table.interrupt_paragraph = False
            token, = block_token.tokenize(lines)
        except ValueError as e:
            raise AssertionError("Token number mismatch.") from e
        finally:
            block_token.Table.interrupt_paragraph = True
        self.assertIsInstance(token, block_token.Paragraph)


class TestTableRow(unittest.TestCase):
    def test_match(self):
        with patch('mistletoe.block_token.TableCell') as mock:
            line = '| cell 1 | cell 2 |\n'
            token = block_token.TableRow(line, line_number=10)
            self.assertEqual(token.row_align, [None])
            mock.assert_has_calls([call('cell 1', None, 10), call('cell 2', None, 10)])

    def test_easy_table_row(self):
        with patch('mistletoe.block_token.TableCell') as mock:
            line = 'cell 1 | cell 2\n'
            token = block_token.TableRow(line, line_number=10)
            self.assertEqual(token.row_align, [None])
            mock.assert_has_calls([call('cell 1', None, 10), call('cell 2', None, 10)])

    def test_short_row(self):
        with patch('mistletoe.block_token.TableCell') as mock:
            line = '| cell 1 |\n'
            token = block_token.TableRow(line, [None, None], 10)
            self.assertEqual(token.row_align, [None, None])
            mock.assert_has_calls([call('cell 1', None, 10), call('', None, 10)])

    def test_escaped_pipe_in_cell(self):
        with patch('mistletoe.block_token.TableCell') as mock:
            line = '| pipe: `\\|` | cell 2\n'
            token = block_token.TableRow(line, line_number=10, row_align=[None, None])
            self.assertEqual(token.row_align, [None, None])
            mock.assert_has_calls([call('pipe: `|`', None, 10), call('cell 2', None, 10)])

    @unittest.skip('Even GitHub fails in here, workaround: always put a space before `|`')
    def test_not_really_escaped_pipe_in_cell(self):
        with patch('mistletoe.block_token.TableCell') as mock:
            line = '|ending with a \\\\|cell 2\n'
            token = block_token.TableRow(line, [None, None], 10)
            self.assertEqual(token.row_align, [None, None])
            mock.assert_has_calls([call('ending with a \\\\', None, 10), call('cell 2', None, 10)])


class TestTableCell(TestToken):
    def test_match(self):
        token = block_token.TableCell('cell 2', line_number=13)
        self._test_token(token, 'cell 2', line_number=13, align=None)


class TestFootnote(unittest.TestCase):
    def test_parse_simple(self):
        lines = ['[key 1]: value1\n',
                 '[key 2]: value2\n']
        token = block_token.Document(lines)
        self.assertEqual(token.footnotes, {"key 1": ("value1", ""),
                                           "key 2": ("value2", "")})

    def test_parse_with_title(self):
        lines = ['[key 1]: value1 "title1"\n',
                 '[key 2]: value2\n',
                 '"title2"\n']
        token = block_token.Document(lines)
        self.assertEqual(token.footnotes, {"key 1": ("value1", "title1"),
                                           "key 2": ("value2", "title2")})

    def test_parse_with_space_in_every_part(self):
        lines = ['[Foo bar]:\n',
                 '<my url>\n',
                 '\'my title\'\n']
        token = block_token.Document(lines)
        self.assertEqual(set(token.footnotes.values()), set({("my url", "my title")}))

    def test_parse_title_must_be_separated_from_link_destination(self):
        lines = ['[foo]: <bar> (baz)\n']
        token = block_token.Document(lines)
        self.assertEqual(set(token.footnotes.values()), set({("bar", "baz")}))

        lines = ['[foo]: <bar>(baz)\n']
        token = block_token.Document(lines)
        self.assertEqual(len(token.footnotes), 0)

    # this tests an edge case, it shouldn't occur in normal documents:
    # "[key 2]" is part of the paragraph above it, because a link reference definitions cannot interrupt a paragraph.
    def test_footnote_followed_by_paragraph(self):
        lines = ['[key 1]: value1\n',
                 'something1\n',
                 '[key 2]: value2\n',
                 'something2\n',
                 '\n',
                 '[key 3]: value3\r\n',  # '\r', or any other whitespace may follow on the same line
                 'something3\n']
        token = block_token.Document(lines)
        self.assertEqual(token.footnotes, {"key 1": ("value1", ""),
                                           "key 3": ("value3", "")})
        self.assertEqual(len(token.children), 2)
        self.assertIsInstance(token.children[0], block_token.Paragraph)
        # children: something1, <line break>, [key 2]: value2, <line break>, something2
        self.assertEqual(len(token.children[0].children), 5)
        self.assertEqual(token.children[0].children[2].content, "[key 2]: value2")
        self.assertEqual(token.children[1].children[0].content, "something3")

    def test_content_after_title_not_allowed(self):
        lines = ['[foo]: /url\n',
                 '"title" ok\n']
        token = block_token.Document(lines)
        self.assertEqual(token.footnotes, {"foo": ("/url", "")})
        self.assertEqual(len(token.children), 1)
        self.assertIsInstance(token.children[0], block_token.Paragraph)
        self.assertEqual(token.children[0].children[0].content, "\"title\" ok")

    def test_footnotes_may_not_have_too_much_leading_space(self):
        lines = ['   [link]: /bla\n',
                 '    [i-am-block-actually]: /foo\n',
                 'paragraph\n',
                 '\n',
                 '\t[i-am-block-too]: /foo\n']
        token = block_token.Document(lines)
        self.assertEqual(token.footnotes, {"link": ("/bla", "")})
        self.assertEqual(len(token.children), 3)
        self.assertIsInstance(token.children[0], block_token.BlockCode)
        self.assertEqual(token.children[0].children[0].content, "[i-am-block-actually]: /foo\n")
        self.assertIsInstance(token.children[1], block_token.Paragraph)
        self.assertEqual(token.children[1].children[0].content, "paragraph")
        self.assertIsInstance(token.children[2], block_token.BlockCode)
        self.assertEqual(token.children[2].children[0].content, "[i-am-block-too]: /foo\n")

    def test_parse_opening_bracket_as_paragraph(self):  # ... and no error is raised
        lines = ['[\n']
        token = block_token.Document(lines)
        self.assertEqual(len(token.footnotes), 0)
        self.assertEqual(len(token.children), 1)

        self.assertIsInstance(token.children[0], block_token.Paragraph)
        self.assertEqual(token.children[0].children[0].content, '[')

    def test_parse_opening_brackets_as_paragraph(self):  # ... and no lines are skipped
        lines = ['[\n',
                 '[ \n',
                 ']\n']
        token = block_token.Document(lines)
        self.assertEqual(len(token.footnotes), 0)
        self.assertEqual(len(token.children), 1)

        para = token.children[0]
        self.assertIsInstance(para, block_token.Paragraph)
        self.assertEqual(len(para.children), 5,
                'expected: RawText, LineBreak, RawText, LineBreak, RawText')
        self.assertEqual(para.children[0].content, '[')


class TestDocument(unittest.TestCase):
    def test_store_footnote(self):
        lines = ['[key 1]: value1\n',
                 '[key 2]: value2\n']
        document = block_token.Document(lines)
        self.assertEqual(document.footnotes['key 1'], ('value1', ''))
        self.assertEqual(document.footnotes['key 2'], ('value2', ''))

    def test_auto_splitlines(self):
        lines = "some\ncontinual\nlines\n"
        document = block_token.Document(lines)
        self.assertIsInstance(document.children[0], block_token.Paragraph)
        self.assertEqual(len(document.children), 1)


class TestThematicBreak(unittest.TestCase):
    def test_match(self):
        def test_case(line):
            token = next(iter(block_token.tokenize([line])))
            self.assertIsInstance(token, block_token.ThematicBreak)
        cases = ['---\n', '* * *\n', '_    _    _\n']
        for case in cases:
            test_case(case)


class TestContains(unittest.TestCase):
    def test_contains(self):
        lines = ['# heading\n', '\n', 'paragraph\n', 'with\n', '`code`\n']
        token = block_token.Document(lines)
        self.assertTrue('heading' in token)
        self.assertTrue('code' in token)
        self.assertFalse('foo' in token)


class TestParent(unittest.TestCase):
    def test_parent(self):
        lines = ['# heading\n', '\n', 'paragraph\n']
        token = block_token.Document(lines)
        self.assertEqual(len(token.children), 2)
        self.assertIsNone(token.parent)
        for child in token.children:
            self.assertEqual(child.parent, token)
            for grandchild in child.children:
                self.assertEqual(grandchild.parent, child)


class TestHtmlBlock(unittest.TestCase):
    def setUp(self):
        block_token.add_token(block_token.HtmlBlock)
        self.addCleanup(block_token.reset_tokens)

    def test_textarea_block_may_contain_blank_lines(self):
        lines = ['<textarea>\n',
                 '\n',
                 '*foo*\n',
                 '\n',
                 '_bar_\n',
                 '\n',
                 '</textarea>\n']
        document = block_token.Document(lines)
        tokens = document.children
        self.assertEqual(1, len(tokens))
        self.assertIsInstance(tokens[0], block_token.HtmlBlock)


class TestLeafBlockTokenContentProperty(unittest.TestCase):
    def setUp(self):
        block_token.add_token(block_token.HtmlBlock)
        self.addCleanup(block_token.reset_tokens)

    def test_code_fence(self):
        lines = ['```\n',
                 'line 1\n',
                 'line 2\n',
                 '```\n']
        document = block_token.Document(lines)
        tokens = document.children
        self.assertEqual(1, len(tokens))
        self.assertIsInstance(tokens[0], block_token.CodeFence)

        # option 1: direct access to the content
        self.assertEqual('line 1\nline 2\n', tokens[0].children[0].content)

        # option 2: using property getter to access the content
        self.assertEqual('line 1\nline 2\n', tokens[0].content)

    def test_block_code(self):
        lines = ['    line 1\n',
                 '    line 2\n']
        document = block_token.Document(lines)
        tokens = document.children
        self.assertEqual(1, len(tokens))
        self.assertIsInstance(tokens[0], block_token.BlockCode)

        # option 1: direct access to the content
        self.assertEqual('line 1\nline 2\n', tokens[0].children[0].content)

        # option 2: using property getter to access the content
        self.assertEqual('line 1\nline 2\n', tokens[0].content)

    def test_html_block(self):
        lines = ['<div>\n',
                 'text\n'
                 '</div>\n']
        document = block_token.Document(lines)
        tokens = document.children
        self.assertEqual(1, len(tokens))
        self.assertIsInstance(tokens[0], block_token.HtmlBlock)

        # option 1: direct access to the content
        self.assertEqual(''.join(lines).strip(), tokens[0].children[0].content)

        # option 2: using property getter to access the content
        self.assertEqual(''.join(lines).strip(), tokens[0].content)


class TestFileWrapper(unittest.TestCase):
    def test_get_set_pos(self):
        lines = [
            "# heading\n",
            "somewhat interesting\n",
            "content\n",
        ]
        wrapper = block_tokenizer.FileWrapper(lines)
        assert next(wrapper) == "# heading\n"
        anchor = wrapper.get_pos()
        assert next(wrapper) == "somewhat interesting\n"
        wrapper.set_pos(anchor)
        assert next(wrapper) == "somewhat interesting\n"

    def test_anchor_reset(self):
        lines = [
            "# heading\n",
            "somewhat interesting\n",
            "content\n",
        ]
        wrapper = block_tokenizer.FileWrapper(lines)
        assert next(wrapper) == "# heading\n"
        wrapper.anchor()
        assert next(wrapper) == "somewhat interesting\n"
        wrapper.reset()
        assert next(wrapper) == "somewhat interesting\n"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys
from importlib import import_module
from time import perf_counter


TEST_FILE = 'test/samples/syntax.md'
TIMES = 1000


def benchmark(package_name):
    def decorator(func):
        def inner():
            try:
                package = import_module(package_name)
            except ImportError:
                return 'not available.'

            start = perf_counter()
            for i in range(TIMES):
                func(package)
            end = perf_counter()

            return end - start
        return inner
    return decorator


@benchmark('markdown')
def run_markdown(package):
    with open(TEST_FILE, 'r', encoding='utf-8') as fin:
        return package.markdown(fin.read(), extensions=['fenced_code', 'tables'])


@benchmark('mistune')
def run_mistune(package):
    with open(TEST_FILE, 'r', encoding='utf-8') as fin:
        return package.markdown(fin.read())


@benchmark('commonmark')
def run_commonmark(package):
    with open(TEST_FILE, 'r', encoding='utf-8') as fin:
        return package.commonmark(fin.read())


@benchmark('mistletoe')
def run_mistletoe(package):
    with open(TEST_FILE, 'r', encoding='utf-8') as fin:
        return package.markdown(fin)


def run(package_name):
    print(package_name, end=': ')
    print(globals()['run_{}'.format(package_name.lower())]())


def run_all(package_names):
    prompt = 'Running tests with {}...'.format(', '.join(package_names))
    print(prompt)
    print('=' * len(prompt))
    for package_name in package_names:
        run(package_name)


def main(*args):
    print('Test document: {}'.format(TEST_FILE))
    print('Test iterations: {}'.format(TIMES))
    if args[1:]:
        run_all(args[1:])
    else:
        run_all(['markdown', 'mistune', 'commonmark', 'mistletoe'])


if __name__ == '__main__':
    main(*sys.argv)



from mistletoe import HtmlRenderer
from pygments import highlight
from pygments.formatters.html import HtmlFormatter
from pygments.lexers import get_lexer_by_name as get_lexer, guess_lexer
from pygments.styles import get_style_by_name as get_style
from pygments.util import ClassNotFound


class PygmentsRenderer(HtmlRenderer):
    formatter = HtmlFormatter()
    formatter.noclasses = True

    def __init__(self, *extras, style='default', fail_on_unsupported_language=False, **kwargs):
        """
        Args:
            extras (list): allows subclasses to add even more custom tokens.
            style (str): short name of the style to be used by Pygments' `HtmlFormatter`,
                      see `pygments.styles.get_style_by_name()`.
            fail_on_unsupported_language (bool): whether to let Pygments' `ClassNotFound`
                      be thrown when there is an unsupported language found on
                      a code block.
                      If `False`, then language is guessed instead of throwing the error.
            **kwargs: additional parameters to be passed to the ancestor's
                      constructor.
        """
        super().__init__(*extras, **kwargs)
        self.formatter.style = get_style(style)
        self.fail_on_unsupported_language = fail_on_unsupported_language

    def render_block_code(self, token):
        code = token.content
        lexer = None

        if token.language:
            try:
                lexer = get_lexer(token.language)
            except ClassNotFound as err:
                if self.fail_on_unsupported_language:
                    raise err

        if lexer is None:
            lexer = guess_lexer(code)

        return highlight(code, lexer, self.formatter)



from unittest import TestCase, mock
from parameterized import parameterized
import mistletoe.latex_renderer
from mistletoe.latex_renderer import LaTeXRenderer
from mistletoe import markdown


class TestLaTeXRenderer(TestCase):
    def setUp(self):
        self.renderer = LaTeXRenderer()
        self.renderer.render_inner = mock.Mock(return_value='inner')
        self.renderer.__enter__()
        self.addCleanup(self.renderer.__exit__, None, None, None)

    def _test_token(self, token_name, expected_output, children=True,
                    without_attrs=None, **kwargs):
        render_func = self.renderer.render_map[token_name]
        children = mock.MagicMock(spec=list) if children else None
        mock_token = mock.Mock(children=children, **kwargs)
        without_attrs = without_attrs or []
        for attr in without_attrs:
            delattr(mock_token, attr)
        self.assertEqual(render_func(mock_token), expected_output)

    def test_strong(self):
        self._test_token('Strong', '\\textbf{inner}')

    def test_emphasis(self):
        self._test_token('Emphasis', '\\textit{inner}')

    def test_inline_code(self):
        func_path = 'mistletoe.latex_renderer.LaTeXRenderer.render_raw_text'

        for content, expected in {'inner': '\\verb|inner|',
                                'a + b': '\\verb|a + b|',
                                'a | b': '\\verb!a | b!',
                                '|ab!|': '\\verb"|ab!|"',
                               }.items():
            with mock.patch(func_path, return_value=content):
                self._test_token('InlineCode', expected, content=content)

        content = mistletoe.latex_renderer.verb_delimiters
        with self.assertRaises(RuntimeError):
            with mock.patch(func_path, return_value=content):
                self._test_token('InlineCode', None, content=content)

    def test_strikethrough(self):
        self._test_token('Strikethrough', '\\sout{inner}')

    def test_image(self):
        expected = '\n\\includegraphics{src}\n'
        self._test_token('Image', expected, src='src')

    @parameterized.expand([
        ('page', '\\href{page}{inner}'),
        ('page%3A+with%3A+escape', '\\href{page\\%3A+with\\%3A+escape}{inner}'),
        ('page#target', '\\href{page\\#target}{inner}')
    ])
    def test_link(self, target, expected):
        self._test_token('Link', expected, target=target)

    @parameterized.expand([
        ('page', '\\url{page}'),
        ('page%3A+with%3A+escape', '\\url{page\\%3A+with\\%3A+escape}'),
        ('page#target', '\\url{page\\#target}')
    ])
    def test_autolink(self, target, expected):
        self._test_token('AutoLink', expected, target=target)

    def test_math(self):
        expected = '$ 1 + 2 = 3 $'
        self._test_token('Math', expected,
                         children=False, content='$ 1 + 2 = 3 $')

    def test_raw_text(self):
        expected = '\\$\\&\\#\\{\\}'
        self._test_token('RawText', expected,
                         children=False, content='$&#{}')

    def test_heading(self):
        expected = '\n\\section{inner}\n'
        self._test_token('Heading', expected, level=1)

    def test_quote(self):
        expected = '\\begin{displayquote}\ninner\\end{displayquote}\n'
        self._test_token('Quote', expected)

    def test_paragraph(self):
        expected = '\ninner\n'
        self._test_token('Paragraph', expected)

    def test_block_code(self):
        func_path = 'mistletoe.latex_renderer.LaTeXRenderer.render_raw_text'
        with mock.patch(func_path, return_value='inner'):
            expected = '\n\\begin{lstlisting}[language=sh]\ninner\\end{lstlisting}\n'
            self._test_token('BlockCode', expected, language='sh')

    def test_list(self):
        expected = '\\begin{itemize}\ninner\\end{itemize}\n'
        self._test_token('List', expected, start=None)

    def test_list_item(self):
        self._test_token('ListItem', '\\item inner\n')

    def test_table_with_header(self):
        func_path = 'mistletoe.latex_renderer.LaTeXRenderer.render_table_row'
        with mock.patch(func_path, autospec=True, return_value='row\n'):
            expected = '\\begin{tabular}{l c r}\nrow\n\\hline\ninner\\end{tabular}\n'
            self._test_token('Table', expected, column_align=[None, 0, 1])

    def test_table_without_header(self):
        expected = ('\\begin{tabular}\ninner\\end{tabular}\n')
        self._test_token('Table', expected, without_attrs=['header'],
                         column_align=[None])

    def test_table_row(self):
        self._test_token('TableRow', ' \\\\\n')

    def test_table_cell(self):
        self._test_token('TableCell', 'inner')

    def test_thematic_break(self):
        self._test_token('ThematicBreak', '\n\\hrulefill\n')

    def test_line_break(self):
        self._test_token('LineBreak', '\\newline\n', soft=False)

    def test_document(self):
        expected = ('\\documentclass{article}\n'
                  '\\begin{document}\n'
                  'inner'
                  '\\end{document}\n')
        self._test_token('Document', expected, footnotes={})


class TestHtmlEntity(TestCase):
    def test_html_entity(self):
        self.assertIn('hello \\& goodbye', markdown('hello &amp; goodbye', LaTeXRenderer))

    def test_html_entity_in_link_target(self):
        self.assertIn('\\href{foo}{hello}', markdown('[hello](f&#111;&#111;)', LaTeXRenderer))


class TestLaTeXFootnotes(TestCase):
    def setUp(self):
        self.renderer = LaTeXRenderer()
        self.renderer.__enter__()
        self.addCleanup(self.renderer.__exit__, None, None, None)

    def test_footnote_image(self):
        from mistletoe import Document
        raw = ['![alt][foo]\n', '\n', '[foo]: bar "title"\n']
        expected = ('\\documentclass{article}\n'
                  '\\usepackage{graphicx}\n'
                  '\\begin{document}\n'
                  '\n'
                  '\n\\includegraphics{bar}\n'
                  '\n'
                  '\\end{document}\n')
        self.assertEqual(self.renderer.render(Document(raw)), expected)

    def test_footnote_link(self):
        from mistletoe import Document
        raw = ['[name][key]\n', '\n', '[key]: target\n']
        expected = ('\\documentclass{article}\n'
                  '\\usepackage{hyperref}\n'
                  '\\begin{document}\n'
                  '\n'
                  '\\href{target}{name}'
                  '\n'
                  '\\end{document}\n')
        self.assertEqual(self.renderer.render(Document(raw)), expected)



from unittest import TestCase
from mistletoe.core_tokens import (MatchObj, Delimiter, follows, shift_whitespace,
        is_control_char, deactivate_delimiters, preceded_by, succeeded_by)


class TestCoreTokens(TestCase):
    def test_match_obj(self):
        match = MatchObj(0, 2, (0, 1, 'a'), (1, 2, 'b'))
        self.assertEqual(match.start(), 0)
        self.assertEqual(match.start(1), 0)
        self.assertEqual(match.start(2), 1)
        self.assertEqual(match.end(), 2)
        self.assertEqual(match.end(1), 1)
        self.assertEqual(match.end(2), 2)
        self.assertEqual(match.group(), 'ab')
        self.assertEqual(match.group(1), 'a')
        self.assertEqual(match.group(2), 'b')

    def test_delimiter(self):
        delimiter = Delimiter(4, 6, 'abcd**')
        self.assertEqual(delimiter.type, '**')
        self.assertEqual(delimiter.number, 2)
        self.assertEqual(delimiter.active, True)
        self.assertEqual(delimiter.start, 4)
        self.assertEqual(delimiter.end, 6)

    def test_delimiter_remove_left(self):
        delimiter = Delimiter(4, 6, 'abcd**')
        self.assertTrue(delimiter.remove(1, left=True))
        self.assertEqual(delimiter.number, 1)
        self.assertEqual(delimiter.start, 5)
        self.assertEqual(delimiter.end, 6)
        self.assertEqual(delimiter.type, '*')

    def test_delimiter_remove_right(self):
        delimiter = Delimiter(4, 6, 'abcd**')
        self.assertTrue(delimiter.remove(1, left=False))
        self.assertEqual(delimiter.number, 1)
        self.assertEqual(delimiter.start, 4)
        self.assertEqual(delimiter.end, 5)
        self.assertEqual(delimiter.type, '*')

    def test_delimiter_remove_right_from_longer_run(self):
        delimiter = Delimiter(4, 8, 'abcd****')
        self.assertTrue(delimiter.remove(1, left=False))
        self.assertEqual(delimiter.number, 3)
        self.assertEqual(delimiter.start, 4)
        self.assertEqual(delimiter.end, 7)
        self.assertEqual(delimiter.type, '***')

    def test_delimiter_remove_empty(self):
        delimiter = Delimiter(4, 6, 'abcd**')
        self.assertFalse(delimiter.remove(2))

    def test_follows(self):
        string = '(foobar)'
        self.assertTrue(follows(string, 6, ')'))
        self.assertFalse(follows(string, 6, '('))
        self.assertFalse(follows(string, 7, ')'))

    def test_shift_whitespace(self):
        string = ' \n\t\rfoo'
        self.assertEqual(shift_whitespace(string, 0), 4)
        self.assertEqual(shift_whitespace('', 0), 0)

    def test_is_control_char(self):
        char = chr(0)
        self.assertTrue(is_control_char(char))
        self.assertFalse(is_control_char('a'))

    def test_deactivate_delimiters(self):
        s = 'abc'
        delimiters = [Delimiter(0, 1, s), Delimiter(1, 2, s), Delimiter(2, 3, s)]
        deactivate_delimiters(delimiters, 2, 'b')
        self.assertTrue(delimiters[0].active)
        self.assertFalse(delimiters[1].active)
        self.assertTrue(delimiters[2].active)

    def test_preceded_by(self):
        whitespace = ' \t\n\r'
        self.assertTrue(preceded_by(1, ' abc', whitespace))
        self.assertTrue(preceded_by(0, 'aabc', whitespace))
        self.assertFalse(preceded_by(1, 'aabc', whitespace))
        self.assertFalse(preceded_by(0, 'aabc', 'abc'))

    def test_succeeded_by(self):
        whitespace = ' \t\n\r'
        self.assertTrue(succeeded_by(3, 'abc ', whitespace))
        self.assertTrue(succeeded_by(4, 'abcc', whitespace))
        self.assertFalse(succeeded_by(3, 'abcc', whitespace))
        self.assertFalse(succeeded_by(4, 'abcc', 'abc'))



import unittest
from mistletoe.span_token import tokenize_inner
from mistletoe.latex_token import Math
from mistletoe.latex_renderer import LaTeXRenderer


class TestLaTeXToken(unittest.TestCase):
    def setUp(self):
        self.renderer = LaTeXRenderer()
        self.renderer.__enter__()
        self.addCleanup(self.renderer.__exit__, None, None, None)

    def test_span(self):
        token = next(iter(tokenize_inner('$ 1 + 2 = 3 $')))
        self.assertIsInstance(token, Math)
        self.assertEqual(token.content, '$ 1 + 2 = 3 $')



"""
GitHub Wiki support for mistletoe.
"""

import re
from mistletoe.span_token import SpanToken
from mistletoe.html_renderer import HtmlRenderer


__all__ = ['GithubWiki', 'GithubWikiRenderer']


class GithubWiki(SpanToken):
    # In order to avoid inefficient backtracking, we use positive lookahead here.
    pattern = re.compile(r"\[\[(?=[^\n|]*\|[^\n]*\]\]) *([^\n|]+?) *\| *([^\n]+?) *\]\]")

    def __init__(self, match):
        self.target = match.group(2)


class GithubWikiRenderer(HtmlRenderer):
    def __init__(self, **kwargs):
        """
        Args:
            **kwargs: additional parameters to be passed to the ancestor's
                      constructor.
        """
        super().__init__(GithubWiki, **kwargs)

    def render_github_wiki(self, token):
        template = '<a href="{target}">{inner}</a>'
        target = self.escape_url(token.target)
        inner = self.render_inner(token)
        return template.format(target=target, inner=inner)



from unittest import TestCase
from unittest.mock import call, patch, sentinel, mock_open, Mock
from mistletoe import cli


class TestCli(TestCase):
    @patch('mistletoe.cli.parse', return_value=Mock(filenames=[], renderer=sentinel.Renderer))
    @patch('mistletoe.cli.interactive')
    def test_main_to_interactive(self, mock_interactive, mock_parse):
        cli.main(None)
        mock_interactive.assert_called_with(sentinel.Renderer)

    @patch('mistletoe.cli.parse', return_value=Mock(filenames=['foo.md'], renderer=sentinel.Renderer))
    @patch('mistletoe.cli.convert')
    def test_main_to_convert(self, mock_convert, mock_parse):
        cli.main(None)
        mock_convert.assert_called_with(['foo.md'], sentinel.Renderer)

    @patch('importlib.import_module', return_value=Mock(Renderer=sentinel.RendererCls))
    def test_parse_renderer(self, mock_import_module):
        namespace = cli.parse(['-r', 'foo.Renderer'])
        mock_import_module.assert_called_with('foo')
        self.assertEqual(namespace.renderer, sentinel.RendererCls)

    def test_parse_filenames(self):
        filenames = ['foo.md', 'bar.md']
        namespace = cli.parse(filenames)
        self.assertEqual(namespace.filenames, filenames)

    @patch('mistletoe.cli.convert_file')
    def test_convert(self, mock_convert_file):
        filenames = ['foo', 'bar']
        cli.convert(filenames, sentinel.RendererCls)
        calls = [call(filename, sentinel.RendererCls) for filename in filenames]
        mock_convert_file.assert_has_calls(calls)

    @patch('mistletoe.markdown', return_value='rendered text')
    @patch('sys.stdout.buffer.write')
    @patch('builtins.open', new_callable=mock_open)
    def test_convert_file_success(self, mock_open_, mock_write, mock_markdown):
        filename = 'foo'
        cli.convert_file(filename, sentinel.RendererCls)
        mock_open_.assert_called_with(filename, 'r', encoding='utf-8')
        mock_write.assert_called_with('rendered text'.encode())

    @patch('builtins.open', side_effect=OSError)
    @patch('sys.exit')
    def test_convert_file_fail(self, mock_exit, mock_open_):
        filename = 'foo'
        cli.convert_file(filename, sentinel.RendererCls)
        mock_open_.assert_called_with(filename, 'r', encoding='utf-8')
        mock_exit.assert_called_with('Cannot open file "foo".')

    @patch('mistletoe.cli._import_readline')
    @patch('mistletoe.cli._print_heading')
    @patch('mistletoe.markdown', return_value='rendered text')
    @patch('builtins.print')
    def test_interactive(self, mock_print, mock_markdown,
                         mock_print_heading, mock_import_readline):
        def MockInputFactory(return_values):
            _counter = -1

            def mock_input(prompt=''):
                nonlocal _counter
                _counter += 1
                if _counter < len(return_values):
                    return return_values[_counter]
                elif _counter == len(return_values):
                    raise EOFError
                else:
                    raise KeyboardInterrupt
            return mock_input

        return_values = ['foo', 'bar', 'baz']
        with patch('builtins.input', MockInputFactory(return_values)):
            cli.interactive(sentinel.RendererCls)

        mock_import_readline.assert_called_with()
        mock_print_heading.assert_called_with(sentinel.RendererCls)
        mock_markdown.assert_called_with(['foo\n', 'bar\n', 'baz\n'],
                sentinel.RendererCls)
        calls = [call('\nrendered text', end=''), call('\nExiting.')]
        mock_print.assert_has_calls(calls)

    @patch('importlib.import_module', return_value=Mock(Renderer=sentinel.RendererCls))
    def test_import_success(self, mock_import_module):
        self.assertEqual(sentinel.RendererCls, cli._import('foo.Renderer'))

    @patch('sys.exit')
    def test_import_incomplete_path(self, mock_exit):
        cli._import('foo')
        error_msg = '[error] please supply full path to your custom renderer.'
        mock_exit.assert_called_with(error_msg)

    @patch('importlib.import_module', side_effect=ImportError)
    @patch('sys.exit')
    def test_import_module_error(self, mock_exit, mock_import_module):
        cli._import('foo.Renderer')
        mock_exit.assert_called_with('[error] cannot import module "foo".')

    @patch('importlib.import_module', return_value=Mock(spec=[]))
    @patch('sys.exit')
    def test_import_class_error(self, mock_exit, mock_import_module):
        cli._import('foo.Renderer')
        error_msg = '[error] cannot find renderer "Renderer" from module "foo".'
        mock_exit.assert_called_with(error_msg)

    @patch('builtins.__import__')
    @patch('builtins.print')
    def test_import_readline_success(self, mock_print, mock_import):
        cli._import_readline()
        mock_print.assert_not_called()

    @patch('builtins.__import__', side_effect=ImportError)
    @patch('builtins.print')
    def test_import_readline_fail(self, mock_print, mock_import):
        cli._import_readline()
        mock_print.assert_called_with('[warning] readline library not available.')

    @patch('builtins.print')
    def test_print_heading(self, mock_print):
        cli._print_heading(Mock(__name__='Renderer'))
        version = cli.mistletoe.__version__
        msgs = ['mistletoe [version {}] (interactive)'.format(version),
                'Type Ctrl-D to complete input, or Ctrl-C to exit.',
                'Using renderer: Renderer']
        calls = [call(msg) for msg in msgs]
        mock_print.assert_has_calls(calls)



"""
Markdown renderer for mistletoe.
"""

import re
from itertools import chain
from typing import Iterable, Optional, Sequence

from mistletoe import block_token, span_token, token
from mistletoe.base_renderer import BaseRenderer


class BlankLine(block_token.BlockToken):
    """
    Blank line token. Represents a single blank line.
    This is a leaf block token without children.
    """

    pattern = re.compile(r"\s*\n$")

    def __init__(self, _):
        self.children = []

    @classmethod
    def start(cls, line):
        return cls.pattern.match(line)

    @classmethod
    def read(cls, lines):
        return [next(lines)]


class LinkReferenceDefinition(span_token.SpanToken):
    """
    Link reference definition. ([label]: dest "title")

    Not included in the parsing process, but called by `LinkReferenceDefinitionBlock`.

    Attributes:
        label (str): link label, used in link references.
        dest (str): link target.
        title (str): link title (default to empty).
    """

    repr_attributes = ("label", "dest", "title")

    def __init__(self, match):
        self.label, self.dest, self.title, self.dest_type, self.title_delimiter = match


class LinkReferenceDefinitionBlock(block_token.Footnote):
    """
    A sequence of link reference definitions.
    This is a leaf block token. Its children are link reference definition tokens.

    This class inherits from `Footnote` and modifies the behavior of the constructor,
    to keep the tokens in the AST.
    """

    def __new__(cls, *args, **kwargs):
        obj = object.__new__(cls)
        obj.__init__(*args, **kwargs)
        return obj

    def __init__(self, matches):
        self.children = [LinkReferenceDefinition(match) for match in matches]


class Fragment:
    """
    Markdown fragment. Used when rendering trees of span tokens into flat sequences.
    May carry additional data in addition to the text.

    Attributes:
        text (str): markdown fragment.
    """

    def __init__(self, text: str, **extras):
        self.text = text
        self.__dict__.update(extras)


class MarkdownRenderer(BaseRenderer):
    """
    Markdown renderer.

    Designed to make as "clean" a roundtrip as possible, markdown -> parsing -> rendering -> markdown,
    except for nonessential whitespace. Except when rendering with word wrapping enabled.

    Includes `HtmlBlock` and `HtmlSpan` tokens in the parsing.
    """

    _whitespace = re.compile(r"\s+")

    def __init__(
        self,
        *extras,
        max_line_length: Optional[int] = None,
        normalize_whitespace: bool = False,
    ):
        """
        Args:
            extras (list): allows subclasses to add even more custom tokens.
            max_line_length (int): if specified, the document is word wrapped to the
                specified line length when rendered. Otherwise the formatting from the
                original (parsed) document is retained as much as possible.
            normalize_whitespace (bool): if `False`, the renderer will try to preserve
                as much whitespace as it currently can. For example, you can
                use this flag to control whether to replace the original
                spacing after every list item leader with just 1 space.
        """
        block_token.remove_token(block_token.Footnote)
        super().__init__(
            *chain(
                (
                    block_token.HtmlBlock,
                    span_token.HtmlSpan,
                    BlankLine,
                    LinkReferenceDefinitionBlock,
                ),
                extras,
            )
        )
        self.render_map["SetextHeading"] = self.render_setext_heading
        self.render_map["CodeFence"] = self.render_fenced_code_block
        self.render_map[
            "LinkReferenceDefinition"
        ] = self.render_link_reference_definition
        self.max_line_length = max_line_length
        self.normalize_whitespace = normalize_whitespace

    def render(self, token: token.Token) -> str:
        """
        Renders the tree of tokens rooted at the given token into markdown.
        """
        if isinstance(token, block_token.BlockToken):
            lines = self.render_map[token.__class__.__name__](
                token, max_line_length=self.max_line_length
            )
        else:
            lines = self.span_to_lines([token], max_line_length=self.max_line_length)

        return "".join(map(lambda line: line + "\n", lines))

    # rendering of span/inline tokens.
    # rendered into sequences of Fragments.

    def render_raw_text(self, token) -> Iterable[Fragment]:
        yield Fragment(token.content, wordwrap=True)

    def render_strong(self, token: span_token.Strong) -> Iterable[Fragment]:
        return self.embed_span(Fragment(token.delimiter * 2), token.children)

    def render_emphasis(self, token: span_token.Emphasis) -> Iterable[Fragment]:
        return self.embed_span(Fragment(token.delimiter), token.children)

    def render_inline_code(self, token: span_token.InlineCode) -> Iterable[Fragment]:
        return self.embed_span(
            Fragment(token.delimiter + token.padding),
            token.children,
            Fragment(token.padding + token.delimiter)
        )

    def render_strikethrough(
        self, token: span_token.Strikethrough
    ) -> Iterable[Fragment]:
        return self.embed_span(Fragment("~~"), token.children)

    def render_image(self, token: span_token.Image) -> Iterable[Fragment]:
        yield Fragment("!")
        yield from self.render_link_or_image(token, token.src)

    def render_link(self, token: span_token.Link) -> Iterable[Fragment]:
        return self.render_link_or_image(token, token.target)

    def render_link_or_image(
        self, token: span_token.SpanToken, target: str
    ) -> Iterable[Fragment]:
        yield from self.embed_span(
            Fragment("["),
            token.children,
            Fragment("]"),
        )

        if token.dest_type == "uri" or token.dest_type == "angle_uri":
            # "[" description "](" dest_part [" " title] ")"
            yield Fragment("(")
            dest_part = "<" + target + ">" if token.dest_type == "angle_uri" else target
            yield Fragment(dest_part)
            if token.title:
                yield from (
                    Fragment(" ", wordwrap=True),
                    Fragment(token.title_delimiter),
                    Fragment(token.title, wordwrap=True),
                    Fragment(
                        ")" if token.title_delimiter == "(" else token.title_delimiter,
                    ),
                )
            yield Fragment(")")
        elif token.dest_type == "full":
            # "[" description "][" label "]"
            yield from (
                Fragment("["),
                Fragment(token.label, wordwrap=True),
                Fragment("]"),
            )
        elif token.dest_type == "collapsed":
            # "[" description "][]"
            yield Fragment("[]")
        else:
            # "[" description "]"
            pass

    def render_auto_link(self, token: span_token.AutoLink) -> Iterable[Fragment]:
        yield Fragment("<" + token.children[0].content + ">")

    def render_escape_sequence(
        self, token: span_token.EscapeSequence
    ) -> Iterable[Fragment]:
        yield Fragment("\\" + token.children[0].content)

    def render_line_break(self, token: span_token.LineBreak) -> Iterable[Fragment]:
        yield Fragment(
            token.content + "\n", wordwrap=token.soft, hard_line_break=not token.soft
        )

    def render_html_span(self, token: span_token.HtmlSpan) -> Iterable[Fragment]:
        yield Fragment(token.content)

    def render_link_reference_definition(
        self, token: LinkReferenceDefinition
    ) -> Iterable[Fragment]:
        yield from (
            Fragment("["),
            Fragment(token.label, wordwrap=True),
            Fragment("]: ", wordwrap=True),
            Fragment(
                "<" + token.dest + ">"
                if token.dest_type == "angle_uri"
                else token.dest,
            ),
        )
        if token.title:
            yield from (
                Fragment(" ", wordwrap=True),
                Fragment(token.title_delimiter),
                Fragment(token.title, wordwrap=True),
                Fragment(
                    ")" if token.title_delimiter == "(" else token.title_delimiter,
                ),
            )

    # rendering of block tokens.
    # rendered into sequences of lines (strings), to be joined by newlines.

    def render_document(
        self, token: block_token.Document, max_line_length: Optional[int]
    ) -> Iterable[str]:
        return self.blocks_to_lines(token.children, max_line_length=max_line_length)

    def render_heading(
        self, token: block_token.Heading, max_line_length: Optional[int]
    ) -> Iterable[str]:
        # note: no word wrapping, because atx headings always fit on a single line.
        line = "#" * token.level
        text = next(self.span_to_lines(token.children, max_line_length=None), "")
        if text:
            line += " " + text
        if token.closing_sequence:
            line += " " + token.closing_sequence
        return [line]

    def render_setext_heading(
        self, token: block_token.SetextHeading, max_line_length: Optional[int]
    ) -> Iterable[str]:
        yield from self.span_to_lines(token.children, max_line_length=max_line_length)
        yield token.underline

    def render_quote(
        self, token: block_token.Quote, max_line_length: Optional[int]
    ) -> Iterable[str]:
        max_child_line_length = max_line_length - 2 if max_line_length else None
        lines = self.blocks_to_lines(
            token.children, max_line_length=max_child_line_length
        )
        return self.prefix_lines(lines or [""], "> ")

    def render_paragraph(
        self, token: block_token.Paragraph, max_line_length: Optional[int]
    ) -> Iterable[str]:
        return self.span_to_lines(token.children, max_line_length=max_line_length)

    def render_block_code(
        self, token: block_token.BlockCode, max_line_length: Optional[int]
    ) -> Iterable[str]:
        lines = token.content[:-1].split("\n")
        return self.prefix_lines(lines, "    ")

    def render_fenced_code_block(
        self, token: block_token.CodeFence, max_line_length: Optional[int]
    ) -> Iterable[str]:
        indentation = " " * token.indentation
        yield indentation + token.delimiter + token.info_string
        yield from self.prefix_lines(
            token.content[:-1].split("\n"), indentation
        )
        yield indentation + token.delimiter

    def render_list(
        self, token: block_token.List, max_line_length: Optional[int]
    ) -> Iterable[str]:
        return self.blocks_to_lines(token.children, max_line_length=max_line_length)

    def render_list_item(
        self, token: block_token.ListItem, max_line_length: Optional[int]
    ) -> Iterable[str]:
        if self.normalize_whitespace:
            prepend = len(token.leader) + 1
            indentation = 0
        else:
            prepend = token.prepend
            indentation = token.indentation
        max_child_line_length = (
            max_line_length - prepend if max_line_length else None
        )
        lines = self.blocks_to_lines(
            token.children, max_line_length=max_child_line_length
        )
        return self.prefix_lines(
            list(lines) or [""],
            " " * indentation + token.leader + " " * (prepend - len(token.leader) - indentation),
            " " * prepend,
        )

    def render_table(
        self, token: block_token.Table, max_line_length: Optional[int]
    ) -> Iterable[str]:
        # note: column widths are not preserved; they are automatically adjusted to fit the contents.
        content = [self.table_row_to_text(token.header), []]
        content.extend(self.table_row_to_text(row) for row in token.children)
        col_widths = self.calculate_table_column_widths(content)
        content[1] = self.table_separator_line_to_text(col_widths, token.column_align)
        return [
            self.table_row_to_line(col_text, col_widths, token.column_align)
            for col_text in content
        ]

    def render_thematic_break(
        self, token: block_token.ThematicBreak, max_line_length: Optional[int]
    ) -> Iterable[str]:
        return [token.line]

    def render_html_block(
        self, token: block_token.HtmlBlock, max_line_length: Optional[int]
    ) -> Iterable[str]:
        return token.content.split("\n")

    def render_link_reference_definition_block(
        self, token: LinkReferenceDefinitionBlock, max_line_length: Optional[int]
    ) -> Iterable[str]:
        # each link reference definition starts on a new line
        for child in token.children:
            yield from self.span_to_lines([child], max_line_length=max_line_length)

    def render_blank_line(
        self, token: BlankLine, max_line_length: Optional[int]
    ) -> Iterable[str]:
        return [""]

    # helper methods

    def embed_span(
        self,
        leader: Fragment,
        tokens: Iterable[span_token.SpanToken],
        trailer: Optional[Fragment] = None,
    ) -> Iterable[Fragment]:
        """
        Makes fragments from `tokens` and embeds within a leader and a trailer.
        The trailer defaults to the same as the leader.
        """
        yield leader
        yield from self.make_fragments(tokens)
        yield trailer or leader

    def blocks_to_lines(
        self, tokens: Iterable[block_token.BlockToken], max_line_length: Optional[int]
    ) -> Iterable[str]:
        """
        Renders a sequence of block tokens into a sequence of lines.
        """
        for token in tokens:  # noqa: F402
            yield from self.render_map[token.__class__.__name__](
                token, max_line_length=max_line_length
            )

    def span_to_lines(
        self, tokens: Iterable[span_token.SpanToken], max_line_length: Optional[int]
    ) -> Iterable[str]:
        """
        Renders a sequence of span (inline) tokens into a sequence of lines.
        """
        fragments = self.make_fragments(tokens)
        return self.fragments_to_lines(fragments, max_line_length=max_line_length)

    def make_fragments(self, tokens: Iterable[span_token.SpanToken]
    ) -> Iterable[Fragment]:
        """
        Renders a sequence of span (inline) tokens into a sequence of Fragments.
        """
        token_list = list(tokens)
        for i, token in enumerate(token_list):
            if (
                isinstance(token, span_token.LineBreak)
                and token.soft
                and (
                    (i + 1 < len(token_list) and isinstance(token_list[i + 1], span_token.HtmlSpan))
                    or (i > 0 and isinstance(token_list[i - 1], span_token.HtmlSpan))
                )
            ):
                yield Fragment("\n", hard_line_break=True)
            else:
                yield from self.render_map[token.__class__.__name__](token)


    @classmethod
    def fragments_to_lines(
        cls, fragments: Iterable[Fragment], max_line_length: Optional[int] = None
    ) -> Iterable[str]:
        """
        Renders a sequence of Fragments into lines.
        With word wrapping, if a `max_line_length` is given, or else following the
        original text flow as closely as possible.
        """
        current_line = ""
        if not max_line_length:
            # plain rendering: merge all fragments and split on newlines
            for fragment in fragments:
                if "\n" in fragment.text:
                    lines = fragment.text.split("\n")
                    yield current_line + lines[0]
                    for inner_line in lines[1:-1]:
                        yield inner_line
                    current_line = lines[-1]
                else:
                    current_line += fragment.text
        else:
            # render with word wrapping
            for word in cls.make_words(fragments):
                if word == "\n":
                    # hard line break
                    yield current_line
                    current_line = ""
                    continue

                if not current_line:
                    # first word on an empty line: accept and continue
                    current_line = word
                    continue

                # try to fit the word on the current line.
                # if it doesn't fit, flush the line and start on the next
                test = current_line + " " + word
                if len(test) <= max_line_length:
                    current_line = test
                else:
                    yield current_line
                    current_line = word

        if current_line:
            yield current_line

    @classmethod
    def make_words(cls, fragments: Iterable[Fragment]) -> Iterable[str]:
        """
        Aggregates and splits a sequence of Fragments into words, i.e., strings
        which do not contain breakable spaces or line breaks. The exception is
        hard line breaks, which are represented by the string `\n`.
        """
        word = ""
        for fragment in fragments:
            if getattr(fragment, "wordwrap", False):
                first = True
                for item in cls._whitespace.split(fragment.text):
                    if first:
                        word += item
                        first = False
                    else:
                        if word:
                            yield word
                        word = item
            elif getattr(fragment, "hard_line_break", False):
                yield from (word + fragment.text[:-1], "\n")
                word = ""
            else:
                word += fragment.text

        if word:
            yield word

    @classmethod
    def prefix_lines(
        cls,
        lines: Iterable[str],
        first_line_prefix: str,
        following_line_prefix: Optional[str] = None,
    ) -> Iterable[str]:
        """
        Prepends a prefix string to a sequence of lines. The first line may
        have a different prefix from the following lines.
        """
        following_line_prefix = following_line_prefix or first_line_prefix
        is_first_line = True
        for line in lines:
            if is_first_line:
                prefixed = first_line_prefix + line
                is_first_line = False
            else:
                prefixed = following_line_prefix + line
            yield prefixed if not prefixed.isspace() else ""

    def table_row_to_text(self, row) -> Sequence[str]:
        """
        Renders each table cell on a table row to text. No word wrapping.
        """
        return [
            self.escape_table_cell_text(
                next(self.span_to_lines(col.children, max_line_length=None), "")
            )
            for col in row.children
        ]

    @staticmethod
    def escape_table_cell_text(text: str) -> str:
        """
        Escapes text that would otherwise be parsed as a table delimiter.
        """
        return text.replace("|", r"\|")

    @classmethod
    def calculate_table_column_widths(cls, col_text) -> Sequence[int]:
        """
        Calculates column widths for a table.
        """
        MINIMUM_COLUMN_WIDTH = 3
        col_widths = []
        for row in col_text:
            while len(col_widths) < len(row):
                col_widths.append(MINIMUM_COLUMN_WIDTH)
            for index, text in enumerate(row):
                col_widths[index] = max(col_widths[index], len(text))
        return col_widths

    @classmethod
    def table_separator_line_to_text(cls, col_widths, col_align) -> Sequence[str]:
        """
        Creates the text for the line separating header from contents in a table
        given column widths and alignments.

        Note: uses dashes for left justified columns, not a colon followed by dashes.
        """
        separator_text = []
        for index, width in enumerate(col_widths):
            align = col_align[index] if index < len(col_align) else None
            sep = ":" if align == 0 else "-"
            sep += "-" * (width - 2)
            sep += ":" if align == 0 or align == 1 else "-"
            separator_text.append(sep)
        return separator_text

    @classmethod
    def table_row_to_line(cls, col_text, col_widths, col_align) -> str:
        """
        Pads/aligns the text for a table row and add the borders (pipe characters).
        """
        padded_text = []
        for index, width in enumerate(col_widths):
            text = col_text[index] if index < len(col_text) else ""
            align = col_align[index] if index < len(col_align) else None
            if align is None:
                padded_text.append("{0: <{w}}".format(text, w=width))
            elif align == 0:
                padded_text.append("{0: ^{w}}".format(text, w=width))
            else:
                padded_text.append("{0: >{w}}".format(text, w=width))
        return "".join(("| ", " | ".join(padded_text), " |"))



"""
Built-in span-level token classes.
"""

import html
import re
import mistletoe.span_tokenizer as tokenizer
from mistletoe import core_tokens, token


"""
Tokens to be included in the parsing process, in the order specified.
"""
__all__ = ['EscapeSequence', 'Strikethrough', 'AutoLink', 'CoreTokens',
           'InlineCode', 'LineBreak', 'RawText']


def tokenize_inner(content):
    """
    A wrapper around span_tokenizer.tokenize. Pass in all span-level token
    constructors as arguments to span_tokenizer.tokenize.

    Doing so (instead of importing span_token module in span_tokenizer)
    avoids cyclic dependency issues, and allows for future injections of
    custom token classes.

    _token_types variable is at the bottom of this module.

    See also: span_tokenizer.tokenize, block_token.tokenize.
    """
    return tokenizer.tokenize(content, _token_types)


def add_token(token_cls, position=1):
    """
    Allows external manipulation of the parsing process.
    This function is called in BaseRenderer.__enter__.

    Arguments:
        token_cls (SpanToken): token to be included in the parsing process.
    """
    _token_types.insert(position, token_cls)


def remove_token(token_cls):
    """
    Allows external manipulation of the parsing process.
    This function is called in BaseRenderer.__exit__.

    Arguments:
        token_cls (SpanToken): token to be removed from the parsing process.
    """
    if token_cls in _token_types:
        _token_types.remove(token_cls)


def reset_tokens():
    """
    Resets global _token_types to all token classes in __all__.
    """
    global _token_types
    _token_types = [globals()[cls_name] for cls_name in __all__]


class SpanToken(token.Token):
    parse_inner = True
    parse_group = 1
    precedence = 5

    def __init__(self, match):
        if not self.parse_inner:
            self.content = match.group(self.parse_group)

    def __contains__(self, text):
        if self.children is not None:
            return any(text in child for child in self.children)
        return text in self.content

    @classmethod
    def find(cls, string):
        return cls.pattern.finditer(string)


class CoreTokens(SpanToken):
    """
    Represents core tokens (Strong, Emphasis, Image, Link) during the early stage of parsing.
    Replaced with objects of the proper classes in the final stage of parsing.
    """
    precedence = 3

    def __new__(self, match):
        return globals()[match.type](match)

    @classmethod
    def find(cls, string):
        return core_tokens.find_core_tokens(string, token._root_node)


class Strong(SpanToken):
    """
    Strong token. ("**some text**")
    This is an inline token. Its children are inline (span) tokens.
    One of the core tokens.
    """
    def __init__(self, match):
        self.delimiter = match.delimiter


class Emphasis(SpanToken):
    """
    Emphasis token. ("*some text*")
    This is an inline token. Its children are inline (span) tokens.
    One of the core tokens.
    """
    def __init__(self, match):
        self.delimiter = match.delimiter


class InlineCode(SpanToken):
    """
    Inline code token. ("`some code`")
    This is an inline token with a single child of type RawText.
    """
    pattern = re.compile(r"(?<!\\|`)(?:\\\\)*(`+)(?!`)(.+?)(?<!`)\1(?!`)", re.DOTALL)
    parse_inner = False
    parse_group = 2

    def __init__(self, match):
        content = match.group(self.parse_group)
        self.delimiter = match.group(1)
        content = content.replace('\n', ' ')
        self.padding = " " if not content.isspace() and content.startswith(" ") and content.endswith(" ") else ""
        if self.padding:
            content = content[1:-1]
        self.children = (RawText(content),)

    @classmethod
    def find(cls, string):
        matches = core_tokens._code_matches
        core_tokens._code_matches = []
        return matches


class Strikethrough(SpanToken):
    """
    Strikethrough token. ("~~some text~~")
    This is an inline token. Its children are inline (span) tokens.
    """
    pattern = re.compile(r"(?<!\\)(?:\\\\)*~~(.+?)~~", re.DOTALL)


class Image(SpanToken):
    """
    Image token. ("![alt](src "title")")
    This is an inline token. Its children are inline (span) tokens holding the image description.
    One of the core tokens.

    Attributes:
        src (str): image source.
        title (str): image title (default to empty).
        label (str): link label, for reference links.
    """
    repr_attributes = ("src", "title")

    def __init__(self, match):
        self.src = EscapeSequence.strip(match.group(2).strip())
        self.title = EscapeSequence.strip(match.group(3))
        self.dest_type = getattr(match, "dest_type", None)
        self.label = getattr(match, "label", None)
        self.title_delimiter = getattr(match, "title_delimiter", None)


class Link(SpanToken):
    """
    Link token. ("[name](target "title")")
    This is an inline token. Its children are inline (span) tokens holding the link text.
    One of the core tokens.

    Attributes:
        target (str): link target.
        title (str): link title (default to empty).
        label (str): link label, for reference links.
    """
    repr_attributes = ("target", "title")

    def __init__(self, match):
        self.target = EscapeSequence.strip(match.group(2).strip())
        self.title = EscapeSequence.strip(match.group(3))
        self.dest_type = getattr(match, "dest_type", None)
        self.label = getattr(match, "label", None)
        self.title_delimiter = getattr(match, "title_delimiter", None)


class AutoLink(SpanToken):
    """
    Autolink token. ("<http://www.google.com>", or "<user@example.org>")
    This is an inline token with a single child of type RawText.

    Attributes:
        children (list): a single RawText node for the link target.
        target (str): link target.
        mailto (bool): true iff the target looks like an email address, but does not have the "mailto:" prefix.
    """
    repr_attributes = ("target", "mailto")
    pattern = re.compile(r"(?<!\\)(?:\\\\)*<(([A-Za-z][A-Za-z0-9+.-]{1,31}):[^ <>]*?|[A-Za-z0-9.!#$%&'*+/=?^_`{|}~-]+@[A-Za-z0-9](?:[A-Za-z0-9-]{0,61}[A-Za-z0-9])?(?:\.[A-Za-z0-9](?:[A-Za-z0-9-]{0,61}[A-Za-z0-9])?)*)>")
    parse_inner = False

    def __init__(self, match):
        content = match.group(self.parse_group)
        scheme = match.group(2)
        self.children = (RawText(content),)
        self.target = content
        self.mailto = not scheme


class EscapeSequence(SpanToken):
    """
    Escape sequence token. ("\\\\*")
    This is an inline token with a single child of type RawText.

    Attributes:
        children (iterator): a single RawText node containing the escaped character.
    """
    pattern = re.compile(r"\\([!\"#$%&'()*+,-./:;<=>?@\[\\\]^_`{|}~])")
    parse_inner = False
    precedence = 2

    def __init__(self, match):
        self.children = (RawText(match.group(self.parse_group)),)

    @classmethod
    def strip(cls, string):
        return html.unescape(cls.pattern.sub(r'\1', string))


class LineBreak(SpanToken):
    """
    Line break token: hard or soft.
    This is an inline token without children.

    Attributes:
        soft (bool): true if this is a soft line break.
    """
    repr_attributes = ("soft",)
    pattern = re.compile(r'( *|\\)\n')
    parse_inner = False
    parse_group = 0

    def __init__(self, match):
        self.content = match.group(1)
        self.soft = not self.content.startswith(('  ', '\\'))


class RawText(SpanToken):
    """
    Raw text token.
    This is an inline token without children.

    RawText is the only token that accepts a string for its constructor,
    instead of a match object. Also, all recursions should bottom out here.
    """
    def __init__(self, content):
        self.content = content


_tags = {'address', 'article', 'aside', 'base', 'basefont', 'blockquote',
        'body', 'caption', 'center', 'col', 'colgroup', 'dd', 'details',
        'dialog', 'dir', 'div', 'dl', 'dt', 'fieldset', 'figcaption', 'figure',
        'footer', 'form', 'frame', 'frameset', 'h1', 'h2', 'h3', 'h4', 'h5',
        'h6', 'head', 'header', 'hr', 'html', 'iframe', 'legend', 'li', 'link',
        'main', 'menu', 'menuitem', 'meta', 'nav', 'noframes', 'ol',
        'optgroup', 'option', 'p', 'param', 'section', 'source', 'summary',
        'table', 'tbody', 'td', 'tfoot', 'th', 'thead', 'title', 'tr', 'track',
        'ul'}

_tag   = r'[A-Za-z][A-Za-z0-9-]*'  # noqa: E221
_attrs = r'(?:\s+[A-Za-z_:][A-Za-z0-9_.:-]*(?:\s*=\s*(?:[^\s"\'=<>`]+|\'[^\']*?\'|"[^\"]*?"))?)*'

_open_tag    = r'(?<!\\)<' + _tag + _attrs + r'\s*/?>'  # noqa: E221
_closing_tag = r'(?<!\\)</' + _tag + r'\s*>'
_comment     = r'(?<!\\)<!--(?!>|->)(?:(?!--).)+?(?<!-)-->'  # noqa: E221
_instruction = r'(?<!\\)<\?.+?\?>'
_declaration = r'(?<!\\)<![A-Z].+?>'
_cdata       = r'(?<!\\)<!\[CDATA.+?\]\]>'  # noqa: E221


class HtmlSpan(SpanToken):
    """
    Span-level HTML token.
    This is an inline token without children.

    Attributes:
        content (str): the raw HTML content.
    """
    pattern = re.compile('|'.join([_open_tag, _closing_tag, _comment,
                                   _instruction, _declaration, _cdata]),
                                   re.DOTALL)
    parse_inner = False
    parse_group = 0


HTMLSpan = HtmlSpan
"""
Deprecated name of the `HtmlSpan` class.
"""


# Note: The following XWiki tokens are based on the XWiki Syntax 2.0 (or above; 1.0 was deprecated years ago already).

class XWikiBlockMacroStart(SpanToken):
    """
    A "block" macro opening tag. ("{{macroName<optionalParams>}}<newLine>")

    We want to keep it on a separate line instead of "soft" merging it with the *following* line.
    """
    pattern = re.compile(r'(?<!\\)(\{\{\w+.*?(?<![\\/])\}\})\s*\n')
    parse_inner = False
    parse_group = 1


class XWikiBlockMacroEnd(SpanToken):
    """
    A "block" macro closing tag. ("<onlySpacesAllowed>{{/macroName}}")

    We want to keep it on a separate line instead of "soft" merging it with the *preceding* line.
    """
    pattern = re.compile(r'^(?:\s*)(\{\{/\w+\}\})', re.MULTILINE)
    parse_inner = False
    parse_group = 1


_token_types = []
reset_tokens()



import unittest
from mistletoe import Document
from mistletoe.contrib.mathjax import MathJaxRenderer


class TestMathJaxRenderer(unittest.TestCase):
    mathjax_src = '<script src="https://cdnjs.cloudflare.com/ajax/libs/mathjax/2.7.0/MathJax.js?config=TeX-MML-AM_CHTML"></script>\n'

    def test_render_html(self):
        with MathJaxRenderer() as renderer:
            token = Document(['# heading 1\n', 'paragraph\n'])
            output = renderer.render(token)
            target = '<h1>heading 1</h1>\n<p>paragraph</p>\n'
            target += self.mathjax_src
            self.assertEqual(output, target)

    def test_render_math(self):
        with MathJaxRenderer() as renderer:
            raw = ['math displayed as a block:\n', '$$ \\sum_{i=1}^{\\infty} \\frac{1}{i^p} $$\n', 'math displayed in-line: $ 2^x $\n']
            token = Document(raw)
            output = renderer.render(token)
            target = '<p>math displayed as a block:\n$$ \\sum_{i=1}^{\\infty} \\frac{1}{i^p} $$\nmath displayed in-line: \\( 2^x \\)</p>\n'
            target += self.mathjax_src
            self.assertEqual(output, target)



import time
from unittest import TestCase, mock
from parameterized import parameterized
from mistletoe import span_token, Document, token
from mistletoe.span_token import tokenize_inner
from mistletoe.contrib.github_wiki import GithubWiki, GithubWikiRenderer


class TestGithubWiki(TestCase):
    def setUp(self):
        token._root_node = Document([])
        self.renderer = GithubWikiRenderer()
        self.renderer.__enter__()
        self.addCleanup(self.renderer.__exit__, None, None, None)

    def test_parse(self):
        MockRawText = mock.Mock()
        RawText = span_token._token_types.pop()
        span_token._token_types.append(MockRawText)
        try:
            tokens = tokenize_inner('text with [[wiki | target]]')
            token = tokens[1]
            self.assertIsInstance(token, GithubWiki)
            self.assertEqual(token.target, 'target')
            MockRawText.assert_has_calls([mock.call('text with '), mock.call('wiki')])
        finally:
            span_token._token_types[-1] = RawText

    def test_render(self):
        token = next(iter(tokenize_inner('[[wiki|target]]')))
        output = '<a href="target">wiki</a>'
        self.assertEqual(self.renderer.render(token), output)

    @parameterized.expand([
        ('[[wiki | target]]', 'wiki', 'target'),
        ('[[ wiki | target ]]', 'wiki', 'target'),
        ('[[a | b | c]]', 'a', 'b | c'),
    ])
    def test_parse_variant_spacing_and_pipes(self, source, label, target):
        token = next(iter(tokenize_inner(source)))
        self.assertIsInstance(token, GithubWiki)
        self.assertEqual(token.target, target)
        self.assertEqual(self.renderer.render_inner(token), label)

    @parameterized.expand([
        ('missing_closing_marker', '[[' + ' ' * 200 + '|' + ' ' * 200 + ']'),
        ('missing_pipe', '[[' + ' ' * 6000 + ']]'),
    ])
    def test_malformed_input_is_rejected_quickly(self, _name, source):
        start = time.perf_counter()
        match = GithubWiki.pattern.match(source)
        elapsed = time.perf_counter() - start
        self.assertLess(elapsed, 1.0)
        self.assertIsNone(match)



from textwrap import dedent
import unittest

from mistletoe import Document
from mistletoe.span_token import Strong
from mistletoe.utils import traverse


class TestTraverse(unittest.TestCase):
    def test(self):
        doc = Document("a **b** c **d**")
        filtered = [t.node.__class__.__name__ for t in traverse(doc)]
        self.assertEqual(
            filtered,
            [
                "Paragraph",
                "RawText",
                "Strong",
                "RawText",
                "Strong",
                "RawText",
                "RawText",
            ],
        )

    def test_with_included_source(self):
        doc = Document(
            dedent(
                """\
            a **b**

            c [*d*](link)
            """
            )
        )
        tree = [
            (
                t.node.__class__.__name__,
                t.parent.__class__.__name__ if t.parent else None,
                t.depth
            )
            for t in traverse(doc, include_source=True)
        ]
        self.assertEqual(
            tree,
            [
                ('Document', None, 0),
                ('Paragraph', 'Document', 1),
                ('Paragraph', 'Document', 1),
                ('RawText', 'Paragraph', 2),
                ('Strong', 'Paragraph', 2),
                ('RawText', 'Paragraph', 2),
                ('Link', 'Paragraph', 2),
                ('RawText', 'Strong', 3),
                ('Emphasis', 'Link', 3),
                ('RawText', 'Emphasis', 4),
            ]
        )

    def test_with_class_filter(self):
        doc = Document("a **b** c **d**")
        filtered = [t.node.__class__.__name__ for t in traverse(doc, klass=Strong)]
        self.assertEqual(filtered, ["Strong", "Strong"])

    def test_with_included_source_and_class_filter(self):
        doc = Document("a **b** c **d**")
        filtered = [
            t.node.__class__.__name__
            for t in traverse(doc, include_source=True, klass=Strong)
        ]
        self.assertEqual(filtered, ["Strong", "Strong"])

    def test_with_depth_limit(self):
        doc = Document("a **b** c **d**")

        # Zero depth with root not included yields no nodes.
        filtered = [t.node.__class__.__name__ for t in traverse(doc, depth=0)]
        self.assertEqual(filtered, [])

        # Zero depth with root included yields the root node.
        filtered = [
            t.node.__class__.__name__
            for t in traverse(doc, depth=0, include_source=True)
        ]
        self.assertEqual(filtered, ["Document"])

        # Depth=1 correctly returns the single node at that level.
        filtered = [t.node.__class__.__name__ for t in traverse(doc, depth=1)]
        self.assertEqual(filtered, ["Paragraph"])

        # Depth=2 returns the correct nodes.
        filtered = [t.node.__class__.__name__ for t in traverse(doc, depth=2)]
        self.assertEqual(
            filtered, ["Paragraph", "RawText", "Strong", "RawText", "Strong"]
        )

        # Depth=3 returns the correct nodes (all nodes in the tree).
        filtered = [t.node.__class__.__name__ for t in traverse(doc, depth=3)]
        self.assertEqual(
            filtered,
            [
                "Paragraph",
                "RawText",
                "Strong",
                "RawText",
                "Strong",
                "RawText",
                "RawText",
            ],
        )

        # Verify there are no additional nodes at depth=4.
        filtered = [t.node.__class__.__name__ for t in traverse(doc, depth=4)]
        self.assertEqual(
            filtered,
            [
                "Paragraph",
                "RawText",
                "Strong",
                "RawText",
                "Strong",
                "RawText",
                "RawText",
            ],
        )



import re
import sys
import json
from mistletoe import Document, HtmlRenderer
from traceback import print_tb
from argparse import ArgumentParser


KNOWN = []
"""
Examples (their numbers) from the specification which are known to fail in mistletoe.
"""


def run_tests(test_entries, start=None, end=None,
              quiet=False, verbose=False, known=False):
    if known:
        print('ignoring tests:', ', '.join(map(str, KNOWN)) + '\n')
    start = start or 0
    end = end or sys.maxsize
    results = [run_test(test_entry, quiet) for test_entry in test_entries
            if test_entry['example'] >= start and test_entry['example'] <= end
            and (not known or test_entry['example'] not in KNOWN)]
    if verbose:
        print_failure_in_sections(results)
    fails = len(list(filter(lambda x: not x[0], results)))
    if fails:
        print('failed:', fails)
        print(' total:', len(results))
    else:
        print('All tests passing.')
    return not fails


def run_test(test_entry, quiet=False):
    test_case = test_entry['markdown'].splitlines(keepends=True)
    try:
        with HtmlRenderer(html_escape_double_quotes=True) as renderer:
            output = renderer.render(Document(test_case))
        success = test_entry['html'] == output
        if not success and not quiet:
            print_test_entry(test_entry, output)
        return success, test_entry['section']
    except Exception as exception:
        if not quiet:
            print_exception(exception, test_entry)
        return False, test_entry['section']


def load_tests(specfile):
    with open(specfile, 'r', encoding='utf-8') as fin:
        tests = json.load(fin)

    # alter the tests for our needs
    for test in tests:
        # CommonMark states that "Renderers may make different decisions about how to escape or normalize URLs in the output.".
        # So we unescape square brackets in URLs to match mistletoe's behavior.
        test['html'] = test['html'].replace('%5B', '[').replace('%5D', ']')

    return tests


def locate_section(section, tests):
    start = None
    end = None
    for test in tests:
        if re.search(section, test['section'], re.IGNORECASE):
            if start is None:
                start = test['example']
        elif start is not None and end is None:
            end = test['example'] - 1
            return start, end
    if start:
        return start, tests[-1]['example'] - 1
    raise RuntimeError("Section '{}' not found, aborting.".format(section))


def print_exception(exception, test_entry):
    print_test_entry(test_entry, '-- exception --', fout=sys.stderr)
    print(exception.__class__.__name__ + ':', exception, file=sys.stderr)
    print('Traceback: ', file=sys.stderr)
    print_tb(exception.__traceback__)


def print_test_entry(test_entry, output, fout=sys.stdout):
    print('example: ', repr(test_entry['example']), file=fout)
    print('markdown:', repr(test_entry['markdown']), file=fout)
    print('html:    ', repr(test_entry['html']), file=fout)
    print('output:  ', repr(output), file=fout)
    print(file=fout)


def print_failure_in_sections(results):
    section = results[0][1]
    failed = 0
    total = 0
    for result in results:
        if section != result[1]:
            if failed:
                section_str = "Failed in section '{}':".format(section)
                result_str = "{:>3} / {:>3}".format(failed, total)
                print('{:70} {}'.format(section_str, result_str))
            section = result[1]
            failed = 0
            total = 0
        if not result[0]:
            failed += 1
        total += 1
    if failed:
        section_str = "Failed in section '{}':".format(section)
        result_str = "{:>3} / {:>3}".format(failed, total)
        print('{:70} {}'.format(section_str, result_str))
    print()


def main():
    parser = ArgumentParser(description="Custom script for running Commonmark tests.")
    parser.add_argument('start', type=int, nargs='?', default=None,
                        help="Run tests starting from this position.")
    parser.add_argument('end', type=int, nargs='?', default=None,
                        help="Run tests until this position.")
    parser.add_argument('-v', '--verbose', dest='verbose', action='store_true',
                        help="Output failure count in every section.")
    parser.add_argument('-q', '--quiet', dest='quiet', action='store_true',
                        help="Suppress failed test entry output.")
    parser.add_argument('-s', '--section', dest='section', default=None,
                        help="Only run tests in specified section.")
    parser.add_argument('-f', '--file', dest='tests', type=load_tests,
                        default='test/specification/commonmark.json',
                        help="Specify alternative specfile to run.")
    parser.add_argument('-n', '--ignore-known', dest='known', action='store_true',
                        help="Ignore tests entries that are known to fail.")
    args = parser.parse_args()

    start = args.start
    end = args.end
    verbose = args.verbose
    quiet = args.quiet
    tests = args.tests
    known = args.known
    if args.section is not None:
        start, end = locate_section(args.section, tests)

    if not run_tests(tests, start, end, quiet, verbose, known):
        sys.exit(1)


if __name__ == '__main__':
    main()



from itertools import chain
from mistletoe import block_token, span_token
from mistletoe.base_renderer import BaseRenderer, URI_SAFE_CHARACTERS


class XWiki20Renderer(BaseRenderer):
    """
    XWiki syntax 2.0 renderer class.

    See mistletoe.base_renderer module for more info.
    """
    def __init__(self, *extras):
        """
        Args:
            extras (list): allows subclasses to add even more custom tokens.
        """
        self.listTokens = []
        self.lastChildOfQuotes = []
        self.firstChildOfListItems = []

        localExtras = [block_token.HtmlBlock, span_token.HtmlSpan, span_token.XWikiBlockMacroStart, span_token.XWikiBlockMacroEnd]
        super().__init__(*chain(localExtras, extras))

    def render_strong(self, token):
        template = '**{}**'
        return template.format(self.render_inner(token))

    def render_emphasis(self, token):
        template = '//{}//'
        return template.format(self.render_inner(token))

    def render_inline_code(self, token):
        # Note: XWiki also offers preformatted text syntax ('##{}##') as a shorter alternative.
        #       We would have to escape the raw text when using it.
        template = '{{{{code}}}}{}{{{{/code}}}}'
        return template.format(self.render_raw_text(token.children[0], False))

    def render_strikethrough(self, token):
        template = '--{}--'
        return template.format(self.render_inner(token))

    def render_image(self, token):
        template = '[[image:{src}]]'
        self.render_inner(token)
        return template.format(src=token.src)

    def render_link(self, token):
        template = '[[{inner}>>{target}]]'
        target = escape_url(token.target)
        inner = self.render_inner(token)
        return template.format(target=target, inner=inner)

    def render_auto_link(self, token):
        template = '[[{target}]]'
        target = escape_url(token.target)
        return template.format(target=target)

    def render_escape_sequence(self, token):
        return '~' + self.render_inner(token)

    def render_raw_text(self, token, escape=True):
        return (token.content.replace('~', '~~')
                # Note: It's probably better to leave potential XWiki macros as-is, i. e. don't escape their markers:
                # .replace('{{', '~{{').replace('}}', '~}}')
                .replace('[[', '~[[').replace(']]', '~]]')
                .replace('**', '~**').replace('//', '~//')
                .replace('##', '~##').replace('--', '~--')
                ) if escape else token.content

    def render_x_wiki_block_macro_start(self, token):
        return token.content + '\n'

    def render_x_wiki_block_macro_end(self, token):
        return '\n' + token.content

    def render_html_span(self, token):
        # XXX: HtmlSpan parses (contains) only individual opening and closing tags
        # => no easy way to wrap the whole HTML code into {{html}} like this:
        #
        # template = '{{{{html wiki="true"}}}}{}{{{{/html}}}}'
        # return template.format(token.content)
        #
        # => Users must do this themselves after the conversion.

        return token.content

    def render_html_block(self, token):
        template = '{{{{html wiki="true"}}}}\n{}\n{{{{/html}}}}' + self._block_eol(token)
        return template.format(token.content)

    def render_heading(self, token):
        template = '{level} {inner} {level}'
        inner = self.render_inner(token)
        return template.format(level='=' * token.level, inner=inner) + self._block_eol(token)

    def render_quote(self, token):
        self.lastChildOfQuotes.append(token.children[-1])
        inner = self.render_inner(token)
        del (self.lastChildOfQuotes[-1])

        return (
            "".join(
                map(
                    lambda line: ">{}{}".format(
                        "" if line.startswith(">") else " ", line
                    ),
                    inner.splitlines(keepends=True),
                )
            )
            + self._block_eol(token)[0:-1]
        )

    def render_paragraph(self, token):
        return '{}'.format(self.render_inner(token)) + self._block_eol(token)

    def render_block_code(self, token):
        template = '{{{{code{attr}}}}}\n{inner}{{{{/code}}}}' + self._block_eol(token)
        if token.language:
            attr = ' language="{}"'.format(token.language)
        else:
            attr = ''

        inner = self.render_raw_text(token.children[0], False)
        return template.format(attr=attr, inner=inner)

    def render_list(self, token):
        inner = self.render_inner(token)
        return inner + self._block_eol(token)[0:-1]

    def render_list_item(self, token):
        template = '{prefix} {inner}\n'
        prefix = ''.join(self.listTokens)
        if '1' in self.listTokens:
            prefix += '.'

        self.firstChildOfListItems.append(token.children[0])
        inner = self.render_inner(token)
        del (self.firstChildOfListItems[-1])

        result = template.format(prefix=prefix, inner=inner.rstrip())

        return result

    def render_inner(self, token):
        if isinstance(token, block_token.List):
            if token.start:
                self.listTokens.append('1')
            else:
                self.listTokens.append('*')

        rendered = [self.render(child) for child in token.children]

        wrap = False
        if isinstance(token, block_token.BlockToken) and len(token.children) > 1:
            # test what follows after the 1st child of this block token - wrap it to a XWiki block right after the 1st child if necessary
            for child in token.children[1:]:
                if isinstance(token, block_token.ListItem) and not isinstance(child, block_token.List):
                    # Note: Nested list within a list item is OK, because it does its own wrapping if necessary.
                    wrap = True
                    break
                if isinstance(token, (block_token.TableCell)) and isinstance(child, block_token.BlockToken):
                    # Note: By-design, Markdown doesn't support multiple lines in one cell, but they can be enforced by using HTML.
                    # See e. g. https://stackoverflow.com/questions/19950648/how-to-write-lists-inside-a-markdown-table.
                    wrap = True
                    break

        if isinstance(token, block_token.List):
            del (self.listTokens[-1])

        return (''.join(rendered) if not wrap
                else '{head}(((\n{tail}\n)))\n'.format(head=rendered[0].rstrip(), tail=''.join(rendered[1:]).rstrip()))

    def render_table(self, token):
        # Copied from JiraRenderer...
        #
        # This is actually gross and I wonder if there's a better way to do it.
        #
        # The primary difficulty seems to be passing down alignment options to
        # reach individual cells.
        template = '{inner}\n'
        if hasattr(token, 'header'):
            head_template = '{inner}'
            header = token.header
            head_inner = self.render_table_row(header, True)
            head_rendered = head_template.format(inner=head_inner)

        else:
            head_rendered = ''

        body_template = '{inner}'
        body_inner = self.render_inner(token)
        body_rendered = body_template.format(inner=body_inner)
        return template.format(inner=head_rendered + body_rendered)

    def render_table_row(self, token, is_header=False):
        if is_header:
            template = '{inner}\n'
        else:
            template = '{inner}\n'

        inner = ''.join([self.render_table_cell(child, is_header)
                         for child in token.children])

        return template.format(inner=inner)

    def render_table_cell(self, token, in_header=False):
        if in_header:
            template = '|={inner}'
        else:
            template = '|{inner}'

        inner = self.render_inner(token)
        return template.format(inner=inner)

    @staticmethod
    def render_thematic_break(token):
        return '----\n'

    @staticmethod
    def render_line_break(token):
        return ' ' if token.soft else '\n'

    def render_document(self, token):
        self.footnotes.update(token.footnotes)
        return self.render_inner(token)

    def _block_eol(self, token):
        return ('\n' if ((len(self.firstChildOfListItems) > 0 and token is self.firstChildOfListItems[-1])
                or (len(self.lastChildOfQuotes) > 0 and token is self.lastChildOfQuotes[-1])) else '\n\n')


def escape_url(raw):
    """
    Escape urls to prevent code injection craziness. (Hopefully.)
    """
    from urllib.parse import quote
    return quote(raw, safe=URI_SAFE_CHARACTERS)



import re
from collections import ChainMap
from functools import reduce
from mistletoe.base_renderer import BaseRenderer
from mistletoe import span_token, block_token
from mistletoe.core_tokens import MatchObj


class Program(block_token.BlockToken):
    def __init__(self, lines):
        self.children = span_token.tokenize_inner(''.join([line.strip() for line in lines]))


class Expr(span_token.SpanToken):
    @classmethod
    def find(cls, string):
        matches = []
        start = []
        for i, c in enumerate(string):
            if c == '(':
                start.append(i)
            elif c == ')':
                pos = start.pop()
                end_pos = i + 1
                content = string[pos + 1:i]
                matches.append(MatchObj(pos, end_pos, (pos + 1, i, content)))
        return matches

    def __repr__(self):
        return '<Expr {}>'.format(self.children)


class Number(span_token.SpanToken):
    pattern = re.compile(r"(\d+)")
    parse_inner = False

    def __init__(self, match):
        self.number = eval(match.group(0))

    def __repr__(self):
        return '<Number {}>'.format(self.number)


class Variable(span_token.SpanToken):
    pattern = re.compile(r"([^\s()]+)")
    parse_inner = False

    def __init__(self, match):
        self.name = match.group(0)

    def __repr__(self):
        return '<Variable {!r}>'.format(self.name)


class Whitespace(span_token.SpanToken):
    parse_inner = False

    def __new__(self, _):
        return None


class Procedure:
    def __init__(self, expr_token, body, env):
        self.params = [child.name for child in expr_token.children]
        self.body = body
        self.env = env


class Scheme(BaseRenderer):
    def __init__(self):
        self.render_map = {
            "Program": self.render_program,
            "Expr": self.render_expr,
            "Number": self.render_number,
            "Variable": self.render_variable,
        }
        block_token._token_types = []
        span_token._token_types = [Expr, Number, Variable, Whitespace]

        self.env = ChainMap({
            "define": self.define,
            "lambda": lambda expr_token, *body: Procedure(expr_token, body, self.env),
            "+": lambda x, y: self.render(x) + self.render(y),
            "-": lambda x, y: self.render(x) - self.render(y),
            "*": lambda x, y: self.render(x) * self.render(y),
            "/": lambda x, y: self.render(x) / self.render(y),
            "<": lambda x, y: self.render(x) < self.render(y),
            ">": lambda x, y: self.render(x) > self.render(y),
            "<=": lambda x, y: self.render(x) <= self.render(y),
            ">=": lambda x, y: self.render(x) >= self.render(y),
            "=": lambda x, y: self.render(x) == self.render(y),
            "true": True,
            "false": False,
            "cons": lambda x, y: (self.render(x), self.render(y)),
            "car": lambda pair: self.render(pair)[0],
            "cdr": lambda pair: self.render(pair)[1],
            "and": lambda *args: all(map(self.render, args)),
            "or": lambda *args: any(map(self.render, args)),
            "not": lambda x: not self.render(x),
            "if": lambda cond, true, false: self.render(true) if self.render(cond) else self.render(false),
            "cond": self.cond,
            "null": None,
            "null?": lambda x: self.render(x) is None,
            "list": lambda *args: reduce(lambda x, y: (y, x), map(self.render, reversed(args)), None),
            "display": lambda *args: print(*map(self.render, args)),
        })

    def render_program(self, token):
        return self.render_inner(token)

    def render_inner(self, token):
        result = None
        for child in token.children:
            result = self.render(child)
        return result

    def render_expr(self, token):
        proc, *args = token.children
        proc = self.render(proc)
        return self.apply(proc, args) if isinstance(proc, Procedure) else proc(*args)

    def render_number(self, token):
        return token.number

    def render_variable(self, token):
        return self.env[token.name]

    def define(self, *args):
        if len(args) == 2:
            name_token, val_token = args
            self.env[name_token.name] = self.render(val_token)
        else:
            name_token, expr_token, *body = args
            self.env[name_token.name] = Procedure(expr_token, body, self.env)

    def cond(self, *exprs):
        for expr in exprs:
            test, value = expr.children
            if test == 'else' and 'else' not in self.env:
                return self.render(value)
            if self.render(test):
                return self.render(value)

    def apply(self, proc, args):
        old_env = self.env
        self.env = proc.env.new_child()
        try:
            for param, arg in zip(proc.params, args):
                self.env[param] = self.render(arg)
            result = None
            for expr in proc.body:
                result = self.render(expr)
        finally:
            self.env = old_env
        return result



from test.base_test import BaseRendererTest
from mistletoe.span_token import tokenize_inner
from mistletoe.contrib.xwiki20_renderer import XWiki20Renderer
import random
import string

filesBasedTest = BaseRendererTest.filesBasedTest


class TestXWiki20Renderer(BaseRendererTest):

    def setUp(self):
        super().setUp()
        self.renderer = XWiki20Renderer()
        self.renderer.__enter__()
        self.addCleanup(self.renderer.__exit__, None, None, None)
        self.sampleOutputExtension = 'xwiki20'

    def genRandomString(self, n, hasWhitespace=False):
        source = string.ascii_letters + string.digits
        if hasWhitespace:
            source = source + ' \t'

        result = ''.join(random.SystemRandom().choice(source) for _ in range(n))
        return result

    def textFormatTest(self, inputTemplate, outputTemplate):
        input = self.genRandomString(80, False)
        token = next(iter(tokenize_inner(inputTemplate.format(input))))
        output = self.renderer.render(token)
        expected = outputTemplate.format(input)
        self.assertEqual(output, expected)

    def test_escaping(self):
        self.textFormatTest('**code: `a = 1;// comment`, plain text URL: http://example.com**',
                '**code: {{{{code}}}}a = 1;// comment{{{{/code}}}}, plain text URL: http:~//example.com**')

    def test_render_strong(self):
        self.textFormatTest('**a{}**', '**a{}**')

    def test_render_emphasis(self):
        self.textFormatTest('*a{}*', '//a{}//')

    def test_render_inline_code(self):
        self.textFormatTest('`a{}b`', '{{{{code}}}}a{}b{{{{/code}}}}')

    def test_render_strikethrough(self):
        self.textFormatTest('~~{}~~', '--{}--')

    def test_render_image(self):
        token = next(iter(tokenize_inner('![image](foo.jpg)')))
        output = self.renderer.render(token)
        expected = '[[image:foo.jpg]]'
        self.assertEqual(output, expected)

    def test_render_link(self):
        url = 'http://{0}.{1}.{2}'.format(self.genRandomString(5), self.genRandomString(5), self.genRandomString(3))
        body = self.genRandomString(80, True)
        token = next(iter(tokenize_inner('[{body}]({url})'.format(url=url, body=body))))
        output = self.renderer.render(token)
        expected = '[[{body}>>{url}]]'.format(url=url, body=body)
        self.assertEqual(output, expected)

    def test_render_auto_link(self):
        url = 'http://{0}.{1}.{2}'.format(self.genRandomString(5), self.genRandomString(5), self.genRandomString(3))
        token = next(iter(tokenize_inner('<{url}>'.format(url=url))))
        output = self.renderer.render(token)
        expected = '[[{url}]]'.format(url=url)
        self.assertEqual(output, expected)

    def test_render_html_span(self):
        markdown = 'text styles: <i>italic</i>, <b>bold</b>'
        # See fixme at the `render_html_span` method...
        # expected = 'text styles: {{html wiki="true"}}<i>italic</i>{{/html}}, {{html wiki="true"}}<b>bold</b>{{/html}}\n\n'
        expected = 'text styles: <i>italic</i>, <b>bold</b>\n\n'
        self.markdownResultTest(markdown, expected)

    def test_render_html_block(self):
        markdown = 'paragraph\n\n<pre>some <i>cool</i> code</pre>'
        expected = 'paragraph\n\n{{html wiki="true"}}\n<pre>some <i>cool</i> code</pre>\n{{/html}}\n\n'
        self.markdownResultTest(markdown, expected)

    def test_render_xwiki_macros_simple(self):
        markdown = """\
{{warning}}
Use this feature with *caution*. See {{Wikipedia article="SomeArticle"/}}. {{test}}Another inline macro{{/test}}.
{{/warning}}
"""
        # Note: There is a trailing ' ' at the end of the second line. It will be a bit complicated to get rid of it.
        expected = """\
{{warning}}
Use this feature with //caution//. See {{Wikipedia article="SomeArticle"/}}. {{test}}Another inline macro{{/test}}. \n\
{{/warning}}

"""
        self.markdownResultTest(markdown, expected)

    def test_render_xwiki_macros_in_list(self):
        markdown = """\
* list item

  {{warning}}
  Use this feature with *caution*. See {{Wikipedia article="SomeArticle"/}}. {{test}}Another inline macro{{/test}}.
  {{/warning}}
"""
        # Note: There is a trailing ' ' at the end of the second line. It will be a bit complicated to get rid of it.
        expected = """\
* list item(((
{{warning}}
Use this feature with //caution//. See {{Wikipedia article="SomeArticle"/}}. {{test}}Another inline macro{{/test}}. \n\
{{/warning}}
)))

"""
        self.markdownResultTest(markdown, expected)

    @filesBasedTest
    def test_render__basic_blocks(self):
        pass

    @filesBasedTest
    def test_render__lists(self):
        pass

    @filesBasedTest
    def test_render__quotes(self):
        pass
